from typing import Callable

from dtoolkit.transformer._util import snake_to_camel
from dtoolkit.transformer.base import MethodTF


def methodtf_factory(
    transform_method: Callable,
    inverse_transform_method: Callable = None,
) -> MethodTF:
    """
    Quickly generate transformer from methods.

    Parameters
    ----------
    transform_method : Callable
        The back algorithm of the :func:`~MethodTF.transform`.

    inverse_transform_method : Callable or None
        The back algorithm of the :func:`~MethodTF.inverse_transform`.

    Examples
    --------
    In your library code::

        from __future__ import annotations

        import numpy as np

        from dtoolkit.transformer.factory import methodtf_factory


        # Generate a plus/minus constant transformer:


        def plus_constant(X: np.ndarray, constant: int | float) -> np.ndarray:
            '''Plus constant to each element of ``X``'''

            return X + constant


        def minus_constant(X: np.ndarray, constant: int | float) -> np.ndarray:
            '''Minus constant to each element of ``X``'''

            return X - constant


        PlusTF = methodtf_factory(plus_constant, minus_constant)


    Back in an interactive IPython session, use this transformer:

    .. code-block:: ipython

        In [1]: import numpy as np

        In [2]: a = np.array([1, 2, 3])

        In [3]: tf = PlusTF(1).update_invargs(1)

        In [4]: tf.transform(a)
        Out[4]:
        [2 3 4]

        In [5]: tf.inverse_transform(a)
        Out[5]:
        [0 1 2]
    """

    if not callable(transform_method):
        raise TypeError("'transform_method' must be a function.")

    if (inverse_transform_method is not None) and (
        not callable(inverse_transform_method)
    ):
        raise TypeError("'inverse_transform_method' must be a function.")

    classname = snake_to_camel(transform_method.__name__) + "TF"

    return type(
        classname,
        (MethodTF,),
        dict(
            transform_method=staticmethod(transform_method),
            inverse_transform_method=staticmethod(inverse_transform_method),
        ),
    )



from dtoolkit.geoaccessor.dataframe.from_wkb import from_wkb  # noqa: F401
from dtoolkit.geoaccessor.dataframe.from_wkt import from_wkt  # noqa: F401
from dtoolkit.geoaccessor.dataframe.from_xy import from_xy  # noqa: F401
from dtoolkit.geoaccessor.dataframe.geocode import geocode  # noqa: F401
from dtoolkit.geoaccessor.dataframe.to_geoframe import to_geoframe  # noqa: F401



from __future__ import annotations

from textwrap import dedent
from typing import Literal
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from pandas.util._decorators import doc
from sklearn.preprocessing import OneHotEncoder as SKOneHotEncoder

from dtoolkit._typing import TwoDimArray
from dtoolkit.accessor.dataframe import cols  # noqa: F401
from dtoolkit.accessor.series import cols  # noqa: F401, F811
from dtoolkit.transformer._compat import SKLEARN_GE_12


if TYPE_CHECKING:
    from scipy.sparse import csr_matrix


class OneHotEncoder(SKOneHotEncoder):
    """
    Encode categorical features as a one-hot numeric array.

    Parameters
    ----------
    categories_with_parent : bool, default False
        Returned column would hook parent labels if ``True`` else
        would be ``categories``.

    sparse : bool, default False
        Will return sparse matrix if ``True`` else will return an array.

    Other parameters
        See :obj:`sklearn.preprocessing.OneHotEncoder`.

    Notes
    -----
    Different to :obj:`sklearn.preprocessing.OneHotEncoder`.
    The result would return a :obj:`~pandas.DataFrame` which uses categories
    as columns.

    Examples
    --------
    Given a dataset with two features, we let the encoder find the unique
    values per feature and transform the data to a binary one-hot encoding.

    :obj:`~pandas.DataFrame` in, :obj:`~pandas.DataFrame` out with categories
    as columns.

    >>> from dtoolkit.transformer import OneHotEncoder
    >>> import pandas as pd
    >>> X = [['Male', 1], ['Female', 3], ['Female', 2]]
    >>> df = pd.DataFrame(X, columns=['gender', 'number'])
    >>> df
        gender  number
    0    Male       1
    1  Female       3
    2  Female       2
    >>> enc = OneHotEncoder()
    >>> enc.fit_transform(df)
       Female  Male    1    2    3
    0     0.0   1.0  1.0  0.0  0.0
    1     1.0   0.0  0.0  0.0  1.0
    2     1.0   0.0  0.0  1.0  0.0

    The encoded data also could hook parent labels.

    >>> enc = OneHotEncoder(categories_with_parent=True)
    >>> enc.fit_transform(df)
       gender_Female  gender_Male  number_1  number_2  number_3
    0            0.0          1.0       1.0       0.0       0.0
    1            1.0          0.0       0.0       0.0       1.0
    2            1.0          0.0       0.0       1.0       0.0
    """

    @doc(SKOneHotEncoder.__init__)
    def __init__(
        self,
        *,
        sparse: bool = False,
        sparse_output: bool = False,
        categories_with_parent: bool = False,
        categories="auto",
        drop=None,
        dtype=np.float64,
        handle_unknown: Literal["error", "ignore", "infrequent_if_exist"] = "error",
        min_frequency: int | float = None,
        max_categories: int = None,
    ):
        # TODO: Remove `sparse` in sklearn 1.4.
        # In the latest (>= 1.1.2) sklearn version, `sparse` is deprecated.
        super().__init__(
            categories=categories,
            drop=drop,
            dtype=dtype,
            handle_unknown=handle_unknown,
            min_frequency=min_frequency,
            max_categories=max_categories,
            **(
                dict(sparse_output=sparse_output)
                if SKLEARN_GE_12
                else dict(sparse=sparse)
            ),
        )
        self.categories_with_parent = categories_with_parent

        # TODO: Remove the following line in sklearn 1.2.
        self.sparse_output = sparse_output

        # compat with sklearn lower version
        # `_parameter_constraints` comes out at sklearn 1.2
        # TODO: delete this condition when required sklearn version is >= 1.2
        if hasattr(self, "_parameter_constraints"):
            self._parameter_constraints["categories_with_parent"] = ["boolean"]

    @doc(
        SKOneHotEncoder.transform,
        dedent(
            """
        Notes
        -----
        This would let :obj:`~pandas.DataFrame` out.
        """,
        ),
    )
    def transform(self, X: TwoDimArray) -> TwoDimArray | csr_matrix:
        from itertools import chain

        Xt = super().transform(X)

        if (
            SKLEARN_GE_12
            and not self.sparse_output
            or not SKLEARN_GE_12
            and not self.sparse
        ) and isinstance(X, (pd.Series, pd.DataFrame)):
            # NOTE: `get_feature_names_out` requires sklearn >= 1.0
            categories = (
                self.get_feature_names_out(X.cols(to_list=True))
                if self.categories_with_parent
                else chain.from_iterable(self.categories_)
            )
            return pd.DataFrame(Xt, columns=categories, index=X.index)

        return Xt



from dtoolkit.accessor.series.bin import bin  # noqa: F401
from dtoolkit.accessor.series.change_axis_type import change_axis_type  # noqa: F401
from dtoolkit.accessor.series.cols import cols  # noqa: F401
from dtoolkit.accessor.series.drop_inf import drop_inf  # noqa: F401
from dtoolkit.accessor.series.dropna_index import dropna_index  # noqa: F401
from dtoolkit.accessor.series.equal import equal  # noqa: F401
from dtoolkit.accessor.series.error_report import error_report  # noqa: F401
from dtoolkit.accessor.series.eval import eval  # noqa: F401
from dtoolkit.accessor.series.expand import expand  # noqa: F401
from dtoolkit.accessor.series.filter_in import filter_in  # noqa: F401
from dtoolkit.accessor.series.getattr import getattr  # noqa: F401
from dtoolkit.accessor.series.groupby_index import groupby_index  # noqa: F401
from dtoolkit.accessor.series.jenks_bin import jenks_bin  # noqa: F401
from dtoolkit.accessor.series.jenks_breaks import jenks_breaks  # noqa: F401
from dtoolkit.accessor.series.len import len  # noqa: F401
from dtoolkit.accessor.series.query import query  # noqa: F401
from dtoolkit.accessor.series.set_unique_index import set_unique_index  # noqa: F401
from dtoolkit.accessor.series.swap_index_values import swap_index_values  # noqa: F401
from dtoolkit.accessor.series.textdistance import textdistance  # noqa: F401
from dtoolkit.accessor.series.textdistance_matrix import (  # noqa: F401
    textdistance_matrix,
)
from dtoolkit.accessor.series.to_datetime import to_datetime  # noqa: F401
from dtoolkit.accessor.series.to_set import to_set  # noqa: F401
from dtoolkit.accessor.series.top_n import top_n  # noqa: F401
from dtoolkit.accessor.series.values_to_dict import values_to_dict  # noqa: F401



from warnings import warn

import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.accessor.register import register_series_method


@register_series_method
@doc(klass="Series", alias="s")
def set_unique_index(s: pd.Series, /, **kwargs) -> pd.Series:
    """
    Set unique index via :meth:`~pandas.{klass}.reset_index`.

    The index will be reset:

    - If ``{alias}.index`` isn't unique.
    - If ``{alias}.index`` isn't monotonic increasing.

    Parameters
    ----------
     **kwargs
        See the documentation for :meth:`~pandas.{klass}.reset_index` for complete
        details on the keyword arguments.

    Returns
    -------
    {klass}
        With new index.

    Warns
    -----
    UserWarning
        If the index of the inputting is not unique.

    See Also
    --------
    pandas.Index.is_monotonic_increasing
    pandas.{klass}.reset_index
    dtoolkit.accessor.series.set_unique_index
    dtoolkit.accessor.dataframe.set_unique_index

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({{'a': [1, 2, 3], 'b': [4, 5, 6]}}, index=[0, 0, 1])
    >>> df
       a  b
    0  1  4
    0  2  5
    1  3  6

    If the index of the inputting is not unique, then it will be reset.
    And give a warning like below::

        UserWarning: The index of 'DataFrame' is not unique.

    >>> df.set_unique_index(drop=True)
       a  b
    0  1  4
    1  2  5
    2  3  6

    The index will be reset if it isn't monotonic increasing or decreasing.

    >>> df = pd.DataFrame({{'a': [1, 2, 3], 'b': [4, 5, 6]}}, index=[3, 1, 2])
    >>> df
       a  b
    3  1  4
    1  2  5
    2  3  6
    >>> df.set_unique_index(drop=True)
       a  b
    0  1  4
    1  2  5
    2  3  6
    """

    if not s.index.is_unique:
        warn(
            f"The 'Index' of {type(s).__name__!r} is not unique.",
            stacklevel=3,
        )
        return s.reset_index(**kwargs)

    elif not s.index.is_monotonic_increasing:
        return s.reset_index(**kwargs)

    return s



import pandas as pd

from dtoolkit.transformer import DataFrameTF
from dtoolkit.transformer._compat import PANDAS_GE_14


class AppendTF(DataFrameTF):
    """
    A transformer could append rows data to the end of caller.

    See Also
    --------
    pandas.DataFrame.append : This transformer's prototype method.

    Examples
    --------
    >>> import pandas as pd
    >>> from dtoolkit.transformer import AppendTF
    >>> df = pd.DataFrame(
    ...     [[1, 2], [3, 4]],
    ...     columns=list("AB"),
    ...     index=["x", "y"],
    ... )
    >>> df
       A  B
    x  1  2
    y  3  4
    >>> df2 = pd.DataFrame(
    ...     [[5, 6], [7, 8]],
    ...     columns=list("AB"),
    ...     index=["x", "y"],
    ... )
    >>> tf = AppendTF(df2)
    >>> tf.transform(df)
       A  B
    x  1  2
    y  3  4
    x  5  6
    y  7  8

    With ``ignore_index`` set to True:

    >>> tf = AppendTF(df2, ignore_index=True)
    >>> tf.transform(df)
       A  B
    0  1  2
    1  3  4
    2  5  6
    3  7  8
    """

    # NOTE: `.append` is deprecated in pandas, and will be removed 2.0.0
    transform_method = staticmethod(
        pd.DataFrame._append if PANDAS_GE_14 else pd.DataFrame.append,
    )



from __future__ import annotations

from typing import Hashable

import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def cols(
    s: pd.Series,
    /,
    to_list: bool = False,
) -> Hashable | None | list[Hashable | None]:
    """
    An API to gather :attr:`~pandas.Series.name` and
    :attr:`~pandas.DataFrame.columns` to one.

    Parameters
    ----------
    to_list : bool, default False
        If True, return a list type.

    Returns
    -------
    Hashable, None or list of them
        The name of the Series.

    See Also
    --------
    pandas.Series.name
    pandas.DataFrame.columns
    dtoolkit.accessor.series.cols
    dtoolkit.accessor.dataframe.cols

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd

    Get :attr:`~pandas.Series.name`.

    >>> s = pd.Series(range(10), name="item")
    >>> s.cols()
    'item'
    >>> s.cols(to_list=True)
    ['item']

    Get :attr:`~pandas.DataFrame.columns`.

    >>> d = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    >>> d.cols()
    ['a', 'b']
    """

    return [s.name] if to_list else s.name



from __future__ import annotations

import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def jenks_breaks(s: pd.Series, /, bins: int) -> list[float]:
    """
    Compute “natural breaks” (Fisher-Jenks algorithm) on Series.

    Parameters
    ----------
    bins : int
        The desired number of class. Requires ``2 <= bins < len(s)``.

    Returns
    -------
    list of floats

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'jenkspy'.

    See Also
    --------
    jenkspy.jenks_breaks
        https://github.com/mthh/jenkspy

    dtoolkit.accessor.series.bin
    dtoolkit.accessor.series.jenks_bin

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([1.3, 7.1, 7.3, 2.3, 3.9, 4.1, 7.8, 1.2, 4.3, 7.3, 5.0, 4.3])
    >>> s
    0     1.3
    1     7.1
    2     7.3
    3     2.3
    4     3.9
    5     4.1
    6     7.8
    7     1.2
    8     4.3
    9     7.3
    10    5.0
    11    4.3
    dtype: float64
    >>> s.jenks_breaks(3)
    [1.2, 2.3, 5.0, 7.8]
    """
    from jenkspy import jenks_breaks

    return jenks_breaks(s, bins)



from __future__ import annotations

from typing import Literal

import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.accessor.dataframe import to_series  # noqa: F401
from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(klass="GeoSeries")
def toposimplify(
    s: gpd.GeoSeries,
    /,
    tolerance: float,
    simplify_algorithm: Literal["dp", "vw"] = "dp",
    simplify_with: Literal["shapely", "simplification"] = "shapely",
    prevent_oversimplify: bool = True,
) -> gpd.GeoSeries:
    """
    Returns a :class:`~geopandas.{klass}` containing a simplified representation
    of each geometry. Similar to :meth:`~geopandas.GeoSeries.simplify`, but keeps
    shared edges.

    .. image:: ../../../../_static/simplify-vs-toposimplify.png
        :width: 80%
        :align: center

    Parameters
    ----------
    tolerance : float
        All parts of a simplified geometry will be no more than tolerance distance from
        the original.

    simplify_algorithm : {{'dp', 'vw'}}, default 'dp'
        ``vw`` will only be selected if ``simplify_with`` is set to ``simplification``.
            - ``dp`` : Douglas-Peucker
            - ``vw`` : Visvalingam-Whyatt

    simplify_with : {{'shapely', 'simplification'}}, default 'shapely'
        Sets the package to use for simplifying. Shapely adopts solely Douglas-Peucker
        and simplification both Douglas-Peucker and Visvalingam-Whyatt. The package
        simplification is known to be quicker than shapely.

    prevent_oversimplify : bool, default True
        If `True`, the simplification is slower, but the likelihood of producing
        valid geometries is higher as it prevents oversimplification. Simplification
        happens on paths separately, so this setting is especially relevant for rings
        with no partial shared paths. This is also known as a topology-preserving
        variant of simplification.

    Returns
    -------
    {klass}

    Raises
    ------
    ModuleNotFoundError
        - If don't have module named 'topojson'.
        - If don't have module named 'simplification'.

    See Also
    --------
    geopandas.GeoSeries.simplify
    dtoolkit.geoaccessor.geoseries.toposimplify
    dtoolkit.geoaccessor.geodataframe.toposimplify
    topojson.Topology.toposimplify
        https://mattijn.github.io/topojson/api/topojson.core.topology.html#toposimplify

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> df = (
    ...     gpd.read_file(gpd.datasets.get_path("naturalearth_lowres"))
    ...     .query('continent == "Africa"')
    ... )
    >>> df.head()
           pop_est  ...                                           geometry
    1   58005463.0  ...  POLYGON ((33.90371 -0.95000, 34.07262 -1.05982...
    2     603253.0  ...  POLYGON ((-8.66559 27.65643, -8.66512 27.58948...
    11  86790567.0  ...  POLYGON ((29.34000 -4.49998, 29.51999 -5.41998...
    12  10192317.3  ...  POLYGON ((41.58513 -1.68325, 40.99300 -0.85829...
    13  52573973.0  ...  POLYGON ((39.20222 -4.67677, 37.76690 -3.67712...
    <BLANKLINE>
    [5 rows x 6 columns]
    >>> df.toposimplify(0.1).head()
           pop_est  ...                                           geometry
    1   58005463.0  ...  POLYGON ((33.90358 -0.94984, 30.76987 -1.01479...
    2     603253.0  ...  POLYGON ((-8.66587 27.65614, -8.81778 27.65614...
    11  86790567.0  ...  POLYGON ((29.34014 -4.49988, 29.27690 -3.29392...
    12  10192317.3  ...  POLYGON ((41.58493 -1.68309, 42.04203 -0.91881...
    13  52573973.0  ...  POLYGON ((39.20182 -4.67669, 39.60462 -4.34688...
    <BLANKLINE>
    [5 rows x 6 columns]
    """
    from topojson import Topology

    return (
        Topology(
            s,
            toposimplify=tolerance,
            simplify_algorithm=simplify_algorithm,
            simplify_with=simplify_with,
            prevent_oversimplify=prevent_oversimplify,
        )
        # `to_gdf` return is a GeoDataFrame, require GeoSeries
        .to_gdf(crs=s.crs).to_series()
        # To fix https://github.com/mattijn/topojson/issues/164
        .set_axis(s.index)
    )



from __future__ import annotations

from typing import Hashable
from typing import Literal
from typing import TYPE_CHECKING

import pandas as pd

from dtoolkit.accessor.register import register_dataframe_method

if TYPE_CHECKING:
    from sklearn.base import RegressorMixin


@register_dataframe_method
def fillna_regression(
    df: pd.DataFrame,
    /,
    method: RegressorMixin,
    columns: dict[Hashable, Hashable | list[Hashable] | pd.Index],
    how: Literal["na", "all"] = "na",
    **kwargs,
) -> pd.DataFrame:
    """
    Fill na value with regression algorithm.

    Parameters
    ----------
    method : RegressorMixin
        Regression transformer.

    columns : dict, ``{y: X}``
        A series of column names pairs. The key is the y (or target) column name, and
        values are X (or feature) column names.

    how : {'na', 'all'}, default 'na'
        Only fill na value or apply regression to entire target column.

    **kwargs
        See the documentation for ``method`` for complete details on
        the keyword arguments.

    Returns
    -------
    DataFrame

    Raises
    ------
    ValueError
        If ``how`` isn't "na" or "all".

    See Also
    --------
    sklearn.kernel_ridge
    sklearn.linear_model
    sklearn.dummy.DummyRegressor
    sklearn.ensemble.AdaBoostRegressor
    sklearn.ensemble.BaggingRegressor
    sklearn.ensemble.ExtraTreesRegressor
    sklearn.ensemble.GradientBoostingRegressor
    sklearn.ensemble.RandomForestRegressor
    sklearn.ensemble.StackingRegressor
    sklearn.ensemble.VotingRegressor
    sklearn.ensemble.HistGradientBoostingRegressor
    sklearn.gaussian_process.GaussianProcessRegressor
    sklearn.isotonic.IsotonicRegression
    sklearn.kernel_ridge.KernelRidge
    sklearn.neighbors.KNeighborsRegressor
    sklearn.neighbors.RadiusNeighborsRegressor
    sklearn.neural_network.MLPRegressor
    sklearn.svm.LinearSVR
    sklearn.svm.NuSVR
    sklearn.svm.SVR
    sklearn.tree.DecisionTreeRegressor
    sklearn.tree.ExtraTreeRegressor

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> from sklearn.linear_model import LinearRegression

    .. math:: y = 1 \\times x_0 + 2 \\times x_1 + 3

    >>> df = pd.DataFrame(
    ...     [
    ...         [1, 1, 6],
    ...         [1, 2, 8],
    ...         [2, 2, 9],
    ...         [2, 3, 11],
    ...         [3, 5, None],
    ...     ],
    ...     columns=['x1', 'x2', 'y'],
    ... )
    >>> df
       x1  x2     y
    0   1   1   6.0
    1   1   2   8.0
    2   2   2   9.0
    3   2   3  11.0
    4   3   5   NaN

    Use 'x1' and 'x2' columns to fit 'y' column and fill the value.

    >>> df.fillna_regression(LinearRegression, {'y': ['x1', 'x2']})
       x1  x2     y
    0   1   1   6.0
    1   1   2   8.0
    2   2   2   9.0
    3   2   3  11.0
    4   3   5  16.0
    """

    if how not in {"na", "all"}:
        raise ValueError(f"invalid inf option: {how!r}")

    df = df.copy()  # avoid mutating the original dataframe
    for y, X in columns.items():
        df = _fillna_regression(df, method, y, X, how=how, **kwargs)

    return df


def _fillna_regression(
    df: pd.DataFrame,
    /,
    method: RegressorMixin,
    y: Hashable,
    X: Hashable | list[Hashable] | pd.Index,
    how: Literal["na", "all"],
    **kwargs,
):
    """Fill single na column at once."""

    if isinstance(X, (str, int)):
        X = [X]

    data = df[df[y].notnull()]
    model = method(**kwargs).fit(data[X], data[y])

    if how == "all":
        df[y] = model.predict(df[X])
    elif how == "na":
        index_null = df[y].isnull()
        df.loc[index_null, y] = model.predict(df.loc[index_null, X])

    return df



from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import filter_geometry as s_filter_geometry
from dtoolkit.geoaccessor.geoseries.filter_geometry import _filter_geometry
from dtoolkit.geoaccessor.geoseries.filter_geometry import BINARY_PREDICATE
from dtoolkit.geoaccessor.register import register_geodataframe_method


if TYPE_CHECKING:
    from shapely.geometry.base import BaseGeometry


@register_geodataframe_method
@doc(s_filter_geometry, klass="GeoDataFrame")
def filter_geometry(
    df: gpd.GeoDataFrame,
    /,
    other: BaseGeometry | gpd.GeoSeries | gpd.GeoDataFrame,
    predicate: BINARY_PREDICATE,
    complement: bool = False,
    **kwargs,
) -> gpd.GeoDataFrame:

    return df[
        _filter_geometry(
            df,
            other=other,
            predicate=predicate,
            complement=complement,
            **kwargs,
        )
    ]



from dtoolkit.accessor.dataframe import filter_in
from dtoolkit.transformer import DataFrameTF


class FilterInTF(DataFrameTF):
    """
    A transformer could filter :obj:`~pandas.DataFrame` contents.

    See Also
    --------
    dtoolkit.accessor.dataframe.filter_in : This transformer's prototype method.

    Examples
    --------
    >>> from dtoolkit.transformer import FilterInTF
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     {
    ...         'legs': [2, 4, 2],
    ...         'wings': [2, 0, 0],
    ...     },
    ...     index=['falcon', 'dog', 'cat'],
    ... )
    >>> df
            legs  wings
    falcon     2      2
    dog        4      0
    cat        2      0

    When ``condition`` is a list check whether every value in the DataFrame is
    present in the list (which animals have 0 or 2 legs or wings).

    Filter rows.

    >>> tf = FilterInTF([0, 2])
    >>> tf.transform(df)
            legs  wings
    falcon     2      2
    cat        2      0

    When ``condition`` is a :obj:`dict`, we can pass values to check for each
    column separately:

    >>> tf = FilterInTF({'legs': [2], 'wings': [2]})
    >>> tf.transform(df)
            legs  wings
    falcon     2      2

    When ``values`` is a Series or DataFrame the index and column must match.
    Note that ``falcon`` does not match based on the number of legs in df2.

    >>> other = pd.DataFrame({'legs': [8, 2], 'wings': [0, 2]},
    ...                      index=['spider', 'falcon'])
    >>> other
            legs  wings
    spider     8      0
    falcon     2      2
    >>> tf = FilterInTF(other)
    >>> tf.transform(df)
            legs  wings
    falcon     2      2
    """

    transform_method = staticmethod(filter_in)



import pandas as pd
from pandas.util._decorators import doc

from dtoolkit._typing import Axis
from dtoolkit.accessor.register import register_series_method


@register_series_method
@doc(alias="s", klass="Series")
def change_axis_type(s: pd.Series, dtype: type, axis: Axis = 0) -> pd.Series:
    """
    Change the type of the axis.

    A sugary syntax wraps :meth:`~pandas.Series.astype`::

        if axis == 0:
            {alias}.index = {alias}.index.astype(dtype)
        else:
            {alias}.columns = {alias}.columns.astype(dtype)

    Parameters
    ----------
    dtype : type
        The type to convert to.

    axis : {{0 or 'index', 1 or 'columns'}}, default 0
        The axis to convert. For Series, always convert index type.

        * 0, or 'index' : Convert index type.
        * 1, or 'columns' : Convert columns type.

    Returns
    -------
    {klass}

    See Also
    --------
    pandas.{klass}.astype
    dtoolkit.accessor.series.change_axis_type
    dtoolkit.accessor.dataframe.change_axis_type

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd

    Convert Series's index type from int to str:

    >>> s = pd.Series(["a", "b"])
    >>> s
    0    a
    1    b
    dtype: object
    >>> s.index
    RangeIndex(start=0, stop=2, step=1)
    >>> s.change_axis_type(str).index
    Index(['0', '1'], dtype='object')

    Convert DataFrame's index type from int to str:

    >>> df = pd.DataFrame([[1, 2], [3, 4]])
    >>> df
       0  1
    0  1  2
    1  3  4
    >>> df.index
    RangeIndex(start=0, stop=2, step=1)
    >>> df.change_axis_type(str).index
    Index(['0', '1'], dtype='object')

    Convert DataFrame's columns type from str to int:

    >>> df.columns
    RangeIndex(start=0, stop=2, step=1)
    >>> df.change_axis_type(str, axis=1).columns
    Index(['0', '1'], dtype='object')
    """

    s = s.copy()  # Avoid mutating the original Series
    s.index = s.index.astype(dtype)

    return s



from __future__ import annotations

from typing import Callable

import pandas as pd
from pandas.api.types import is_string_dtype

from dtoolkit.accessor.register import register_series_method
from dtoolkit.accessor.series.textdistance import textdistance


@register_series_method
def textdistance_matrix(
    s: pd.Series,
    /,
    other: None | pd.Series = None,
    method: Callable = None,
) -> pd.DataFrame:
    """
    Returns a ``DataFrame`` containing the text distances matrix between in ``s``
    and ``other``.

    Parameters
    ----------
    other : None or Series, default None
        If None, use ``s``.

    method : Callable, default None
        The method to calculate the distance. The first and second positional parameters
        will be compared. If None, \
`rapidfuzz.fuzz.ratio <https://maxbachmann.github.io/RapidFuzz/Usage/fuzz.html#ratio>`_.

    Returns
    -------
    DataFrame
        The values are the text distances.

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'rapidfuzz'.

    TypeError
        - If ``s`` is not string dtype.
        - If ``other`` is not string dtype.

    See Also
    --------
    textdistance

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series(["hello", "world"])
    >>> s
    0    hello
    1    world
    dtype: object
    >>> s.textdistance_matrix(pd.Series(["hello", "python"]))
           0          1
    0  100.0  36.363636
    1   20.0  18.181818
    """

    if other is None:
        other = s.copy()
    if not isinstance(other, pd.Series):
        raise TypeError(f"Expected Series(string), but got {type(other).__name__!r}.")
    if not is_string_dtype(other):
        raise TypeError(f"Expected Series(string), but got {other.dtype!r}.")

    return pd.concat(
        (textdistance(s, o, method=method) for o in other),
        axis=1,
        keys=other.index,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
import pandas as pd
from geopandas.base import is_geometry_type

from dtoolkit.accessor.register import register_series_method

if TYPE_CHECKING:
    from pyproj import CRS


@register_series_method
def to_geoframe(
    s: pd.Series,
    /,
    geometry: gpd.GeoSeries = None,
    crs: CRS | str | int = None,
    **kwargs,
) -> gpd.GeoDataFrame | pd.DataFrame:
    """
    Transform an array of shapely scalars :class:`~pandas.Series` to
    a :class:`~geopandas.GeoDataFrame`.

    Parameters
    ----------
    geometry : GeoSeries, optional
        It will be prior set as 'geometry' column on GeoDataFrame.

    crs : CRS, str, int, optional
        Coordinate Reference System of the geometry objects. Can be anything
        accepted by :meth:`~pyproj.crs.CRS.from_user_input`, such as an authority
        string (eg "EPSG:4326" / 4326) or a WKT string.

    **kwargs
        See the documentation for :class:`~geopandas.GeoDataFrame` and  for complete
        details on the keyword arguments.

    Returns
    -------
    DataFrame or GeoDataFrame
        GeoDataFrame if the data is an array of shapely scalars or ``geometry`` is set.

    See Also
    --------
    dtoolkit.geoaccessor.series.to_geoseries
    dtoolkit.geoaccessor.dataframe.to_geoframe

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> s = pd.Series(
    ...     pd.Series(
    ...         [
    ...             "POINT (1 1)",
    ...             "POINT (2 2)",
    ...             "POINT (3 3)",
    ...         ],
    ...     )
    ...     .from_wkt(drop=True, crs=4326)
    ... )
    >>> s
    0    POINT (1.00000 1.00000)
    1    POINT (2.00000 2.00000)
    2    POINT (3.00000 3.00000)
    dtype: geometry
    >>> type(s)
    <class 'pandas.core.series.Series'>
    >>> gs = s.to_geoframe()
    >>> gs
                        geometry
    0    POINT (1.00000 1.00000)
    1    POINT (2.00000 2.00000)
    2    POINT (3.00000 3.00000)
    >>> type(gs)
    <class 'geopandas.geodataframe.GeoDataFrame'>
    """

    if geometry is not None:
        return gpd.GeoDataFrame(s, geometry=geometry, crs=crs, **kwargs)
    elif is_geometry_type(s):
        return gpd.GeoDataFrame(geometry=s, crs=crs, **kwargs)
    else:
        return s.to_frame()



import geopandas as gpd
import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def geocode(s: pd.Series, /, drop: bool = False, **kwargs) -> gpd.GeoDataFrame:
    """
    Geocode string type Series and get a GeoDataFrame of the resulting points.

    Parameters
    ----------
    drop : bool, default False
        Don't contain the original data anymore.

    **kwargs
        See the documentation for :func:`~geopandas.tools.geocode` for complete details
        on the keyword arguments.

    Returns
    -------
    GeoDataFrame

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'geopy'.

    ValueError
        If 'drop' is True and the name of Series is empty.

    See Also
    --------
    geopandas.tools.geocode
    dtoolkit.geoaccessor.dataframe.geocode

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> s = pd.Series(
    ...     [
    ...         "boston, ma",
    ...         "1600 pennsylvania ave. washington, dc",
    ...     ],
    ... )
    >>> s
    0                               boston, ma
    1    1600 pennsylvania ave. washington, dc
    dtype: object
    >>> s.geocode(drop=True)
                         geometry                                            address
    0  POINT (-71.06051 42.35543)               Boston, Massachusetts, United States
    1  POINT (-77.03655 38.89770)  White House, 1600, Pennsylvania Avenue Northwe...
    """

    if not drop and s.name is None:
        raise ValueError(
            "to keep the original data requires setting the 'name' of "
            f"{s.__class__.__name__!r}",
        )

    df = gpd.tools.geocode(s, **kwargs)
    return df if drop else pd.concat((df, s), axis=1)



from __future__ import annotations

from functools import wraps
from typing import Callable

from pandas.api.extensions import register_dataframe_accessor
from pandas.api.extensions import register_index_accessor
from pandas.api.extensions import register_series_accessor
from pandas.util._decorators import doc

from dtoolkit._typing import SeriesOrFrame


def register_method_factory(register_accessor, /):
    """
    Let pandas-object like accessor which only hooks class also hooks function easily.

    Read more in the `User Guide`_.

    .. _User Guide: ../../guide/tips_about_accessor.ipynb#Extend-to-Pandas-like-Object

    Parameters
    ----------
    register_accessor : Pandas-object like accessor

    See Also
    --------
    register_dataframe_method
    register_series_method
    register_index_method
    dtoolkit.geoaccessor.register_geoseries_method
    dtoolkit.geoaccessor.register_geodataframe_method
    """

    # based on pandas_flavor/register.py
    def register_accessor_method(method: Callable, name: str, /):
        @wraps(method)
        def method_accessor(pd_obj: SeriesOrFrame, /):
            @wraps(method)
            def wrapper(*args, **kwargs):
                return method(pd_obj, *args, **kwargs)

            return wrapper

        # Register method as pandas object inner method.
        register_accessor(name)(method_accessor)

        # Must return method itself, otherwise would get None.
        return method

    def register_accessor_alias(name: str = None, /):
        def wrapper(method: Callable, /):
            return register_accessor_method(method, name or method.__name__)

        return wrapper

    @wraps(register_accessor)
    def decorator(name: Callable | str = None, /):
        if callable(name):  # Supports `@register_*_method` using.
            method = name  # This 'name' variable actually is a function.
            return register_accessor_method(method, method.__name__)

        # Supports `@register_*_method()` using.
        return register_accessor_alias(name)

    return decorator


@register_method_factory
@doc(klass=":class:`~pandas.Series`")
def register_series_method(name: str = None):
    """
    {klass} register accessor for human.

    Write method normally, use method naturally.

    Read more in the `User Guide`_.

    .. _User Guide: ../../guide/tips_about_accessor.ipynb

    Parameters
    ----------
    name : str, optional
        Use the ``method`` name as the default accessor entrance if ``name`` is None.

    See Also
    --------
    register_dataframe_method
    register_series_method
    register_index_method
    pandas.api.extensions.register_dataframe_accessor
    pandas.api.extensions.register_series_accessor
    pandas.api.extensions.register_index_accessor

    Examples
    --------
    In your library code::

        from __future__ import annotations

        from dtoolkit.accessor import register_dataframe_method
        from dtoolkit.accessor import register_series_method
        from dtoolkit.accessor import register_index_method
        import pandas as pd

        @register_index_method("col")  # Support alias name also.
        @register_series_method("col")
        @register_dataframe_method("col")
        @register_index_method  # Use accessor method's `__name__` as the entrance.
        @register_series_method
        @register_dataframe_method
        def cols(pd_obj) -> int | str | list[int | str] | None:
            '''
            An API to gather :attr:`~pandas.Series.name` and
            :attr:`~pandas.DataFrame.columns` to one.
            '''

            if isinstance(pd_obj, (pd.Series, pd.Index)):
                return pd_obj.name

            return pd_obj.columns.tolist()

    Back in an interactive IPython session:

    .. code-block:: ipython

        In [1]: import pandas as pd

        In [2]: df = pd.DataFrame(
           ...:     {{
           ...:         "a": [1, 2],
           ...:         "b": [3, 4],
           ...:     }},
           ...:     index=pd.Index(
           ...:         ["x", "y"],
           ...:         name="c",
           ...:     ),
           ...: )

        In [3]: df
        Out[3]:
           a  b
        c
        x  1  3
        y  2  4

        Get the columns of DataFrame via `cols` or `col` method

        In [4]: df.col()
        Out[4]: ['a', 'b']

        Get name of Series via `cols` or `col` method

        In [5]: df.a.col()
        Out[5]: 'a'

        Get name of Index via `cols` or `col` method

        In [6]: df.index.col()
        Out[6]: 'c'
    """

    return register_series_accessor(name)


@register_method_factory
@doc(register_series_method, klass=":class:`~pandas.DataFrame`")
def register_dataframe_method(name: str = None):
    return register_dataframe_accessor(name)


@register_method_factory
@doc(register_series_method, klass=":class:`~pandas.Index`")
def register_index_method(name: str = None):
    return register_index_accessor(name)



import geopandas as gpd
import pandas as pd

from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
def drop_geometry(df: gpd.GeoDataFrame, /) -> pd.DataFrame:
    """
    Drop the activate geometry column from the :class:`~geopandas.GeoDataFrame`
    to return a normal :class:`~pandas.DataFrame`.

    Returns
    -------
    DataFrame
        Dropped active geometry column.

    See Also
    --------
    geopandas.GeoDataFrame.rename_geometry
    geopandas.GeoDataFrame.set_geometry

    Notes
    -----
    If you want to set another column as the active geometry, please use
    :meth:`~geopandas.GeoDataFrame.set_geometry` instead.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     {
    ...         "x": [122, 100],
    ...         "y": [55, 1],
    ...         "name": ["a", "b"],
    ...     },
    ... ).from_xy("x", "y", crs=4326)
    >>> df
         x   y name                    geometry
    0  122  55    a  POINT (122.00000 55.00000)
    1  100   1    b   POINT (100.00000 1.00000)
    >>> df.drop_geometry()
         x   y name
    0  122  55    a
    1  100   1    b

    Drop customed geometry column.

    >>> df_new = df.rename_geometry("geom")
    >>> df_new
         x   y name                        geom
    0  122  55    a  POINT (122.00000 55.00000)
    1  100   1    b   POINT (100.00000 1.00000)
    >>> df_new.drop_geometry()
         x   y name
    0  122  55    a
    1  100   1    b
    """

    result = df.drop(columns=df.geometry.name)

    # The return of `.drop(columns='geometry')` is a `GeoDataFrame`
    # not a `DataFrame`in geopandas version from 0.9.0 to 0.10.2.
    # So there can't return directly.
    # TODO: Delete this code block when the required minimal version
    # of geopandas is 0.11.0.
    return pd.DataFrame(result) if isinstance(result, gpd.GeoDataFrame) else result



import pandas as pd

from dtoolkit.transformer import DataFrameTF


class EvalTF(DataFrameTF):
    """
    A transformer could evaluate a string describing operations on
    :obj:`~pandas.DataFrame` columns.

    See Also
    --------
    pandas.DataFrame.eval : This transformer's prototype method.

    Notes
    -----
    :meth:`~pandas.DataFrame.eval`'s ``inplace`` parameter is not work for
    this transformer. Actually this break pipeline stream. If a transformer's
    ``inplace`` is ``True``, the next tf input would get ``None``.

    Examples
    --------
    >>> import pandas as pd
    >>> from dtoolkit.transformer import EvalTF
    >>> df = pd.DataFrame({'A': range(1, 6), 'B': range(10, 0, -2)})
    >>> df
        A   B
    0  1  10
    1  2   8
    2  3   6
    3  4   4
    4  5   2
    >>> tf = EvalTF('A + B')
    >>> tf.transform(df)
    0    11
    1    10
    2     9
    3     8
    4     7
    dtype: int64

    Assignment is allowed though by default the original DataFrame is not
    modified.

    >>> tf = EvalTF('C = A + B')
    >>> tf.transform(df)
        A   B   C
    0  1  10  11
    1  2   8  10
    2  3   6   9
    3  4   4   8
    4  5   2   7
    >>> df
        A   B
    0  1  10
    1  2   8
    2  3   6
    3  4   4
    4  5   2

    Multiple columns can be assigned to using multi-line expressions:

    >>> tf = EvalTF(
    ...     '''
    ... C = A + B
    ... D = A - B
    ... '''
    ... )
    >>> tf.transform(df)
        A   B   C  D
    0  1  10  11 -9
    1  2   8  10 -6
    2  3   6   9 -3
    3  4   4   8  0
    4  5   2   7  3
    """

    transform_method = staticmethod(pd.DataFrame.eval)



import pandas as pd

from dtoolkit.accessor.register import register_series_method
from dtoolkit.accessor.series.jenks_breaks import jenks_breaks


@register_series_method("jenks_cut")
@register_series_method
def jenks_bin(s: pd.Series, /, bins: int, **kwargs) -> pd.Series:
    """
    Bin values into discrete intervals via “natural breaks” (Fisher-Jenks algorithm).

    Parameters
    ----------
    bins : int
        The desired number of class. Requires ``2 <= bins < len(s)``.

    **kwargs
        See the documentation for :meth:`pandas.cut` for complete details on
        the keyword arguments.

    Returns
    -------
    Series

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'jenkspy'.

    See Also
    --------
    dtoolkit.accessor.series.bin
    dtoolkit.accessor.series.jenks_breaks

    Notes
    -----
    This method could be called via ``s.jenks_bin`` or ``df.jenks_cut``.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([1.3, 7.1, 7.3, 2.3, 3.9, 4.1, 7.8, 1.2, 4.3, 7.3, 5.0, 4.3])
    >>> s
    0     1.3
    1     7.1
    2     7.3
    3     2.3
    4     3.9
    5     4.1
    6     7.8
    7     1.2
    8     4.3
    9     7.3
    10    5.0
    11    4.3
    dtype: float64
    >>> s.jenks_bin(3, include_lowest=True)
    0     (1.199, 2.3]
    1       (5.0, 7.8]
    2       (5.0, 7.8]
    3     (1.199, 2.3]
    4       (2.3, 5.0]
    5       (2.3, 5.0]
    6       (5.0, 7.8]
    7     (1.199, 2.3]
    8       (2.3, 5.0]
    9       (5.0, 7.8]
    10      (2.3, 5.0]
    11      (2.3, 5.0]
    dtype: category
    Categories (3, interval[float64, right]): [(1.199, 2.3] < (2.3, 5.0] < (5.0, 7.8]]
    """

    bins = jenks_breaks(s, bins)
    return pd.cut(s, bins=bins, **kwargs)



from dtoolkit.accessor.dataframe.boolean import boolean  # noqa: F401
from dtoolkit.accessor.dataframe.change_axis_type import change_axis_type  # noqa: F401
from dtoolkit.accessor.dataframe.cols import cols  # noqa: F401
from dtoolkit.accessor.dataframe.decompose import decompose  # noqa: F401
from dtoolkit.accessor.dataframe.drop_inf import drop_inf  # noqa: F401
from dtoolkit.accessor.dataframe.drop_or_not import drop_or_not  # noqa: F401
from dtoolkit.accessor.dataframe.dropna_index import dropna_index  # noqa: F401
from dtoolkit.accessor.dataframe.equal import equal  # noqa: F401
from dtoolkit.accessor.dataframe.expand import expand  # noqa: F401
from dtoolkit.accessor.dataframe.fillna_regression import (  # noqa: F401
    fillna_regression,
)
from dtoolkit.accessor.dataframe.filter_in import filter_in  # noqa: F401
from dtoolkit.accessor.dataframe.groupby_index import groupby_index  # noqa: F401
from dtoolkit.accessor.dataframe.repeat import repeat  # noqa: F401
from dtoolkit.accessor.dataframe.set_unique_index import set_unique_index  # noqa: F401
from dtoolkit.accessor.dataframe.to_series import to_series  # noqa: F401
from dtoolkit.accessor.dataframe.top_n import top_n  # noqa: F401
from dtoolkit.accessor.dataframe.values_to_dict import values_to_dict  # noqa: F401
from dtoolkit.accessor.dataframe.weighted_mean import weighted_mean  # noqa: F401



from warnings import warn

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from dtoolkit.accessor.register import register_series_method
from dtoolkit.util._exception import find_stack_level


@register_series_method
def equal(s: pd.Series, /, other, align: bool = True, **kwargs) -> pd.Series:
    """
    Return a boolean Series containing the result of comparing with ``other``.

    A sugar syntax for ``np.equal(s, other, **kwargs)``.

    Parameters
    ----------
    other : scalar or 1d array-like
        The value(s) to compare with the Series.

        - scalar : compare each element with the scalar value.
        - 1d array-like : compare each element with the corresponding value.

    align : bool, default True
        If True, align ``s`` and ``other`` before comparing. Only works while
        ``other`` is a Series.

    **kwargs
        Additional keyword arguments are passed to :meth:`numpy.equal`.

    Returns
    -------
    Series

    Warns
    -----
    UserWarning
        If ``align`` is True and ``other`` is a Series, but its indexes are not
        equal to ``s``.

    Raises
    ------
    ValueError
        If the lengths of ``s`` and ``other`` are not equal.

    See Also
    --------
    pandas.Series.eq : Compare two Series.
    numpy.equal : Return (x1 == x2) element-wise.
    dtoolkit.accessor.dataframe.equal : DataFrame version of ``.equal``.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([1, 2, 3])
    >>> s
    0    1
    1    2
    2    3
    dtype: int64

    Like ``Series.eq`` method to compare scalar or array-like via ``==``,
    ``Series.equal`` could also do that.

    Compare with scalar.

    >>> s == 2
    0    False
    1     True
    2    False
    dtype: bool
    >>> s.equal(2)
    0    False
    1     True
    2    False
    dtype: bool

    Compare with array-like.

    >>> s.equal(pd.Series([3, 2, 1]))
    0    False
    1     True
    2    False
    dtype: bool
    >>> s.equal(pd.Series([3, 2, 1]))
    0    False
    1     True
    2    False
    dtype: bool

    ``s.equal`` can also compare with array-like based on index.

    >>> s == pd.Series([3, 2, 1], index=[2, 1, 0])  # doctest: +SKIP
    ValueError: Can only compare identically-labeled Series objects
    >>> s.equal(pd.Series([3, 2, 1], index=[2, 1, 0]))
    0    True
    1    True
    2    True
    dtype: bool
    >>> s.equal(pd.Series([3, 2, 1], index=[2, 1, 0]), align=False)
    0    False
    1     True
    2    False
    dtype: bool
    """

    if align and isinstance(other, pd.Series) and not s.index.equals(other.index):
        warn("the indices are different.", stacklevel=find_stack_level())
        s, other = s.align(other)

    if is_list_like(other):
        other = np.asarray(other)

    return np.equal(s, other, **kwargs)



import pandas as pd

from dtoolkit.transformer import DataFrameTF


class DropTF(DataFrameTF):
    """
    A transformer could drop specified labels from rows or columns.

    See Also
    --------
    pandas.DataFrame.drop : This transformer's prototype method.

    Notes
    -----
    :meth:`~pandas.DataFrame.drop`'s ``inplace`` parameter is not work for
    this transformer. Actually this break pipeline stream. If a transformer's
    ``inplace`` is ``True``, the next tf input would get ``None``.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dtoolkit.transformer import DropTF
    >>> df = pd.DataFrame(np.arange(12).reshape(3, 4),
    ...                   columns=['A', 'B', 'C', 'D'])
    >>> df
        A  B   C   D
    0  0  1   2   3
    1  4  5   6   7
    2  8  9  10  11

    Drop columns

    >>> tf = DropTF(['B', 'C'], axis=1)
    >>> tf.transform(df)
        A   D
    0  0   3
    1  4   7
    2  8  11

    Drop a row by index

    >>> tf = DropTF([0, 1])
    >>> tf.transform(df)
        A  B   C   D
    2  8  9  10  11

    Drop columns and/or rows of MultiIndex DataFrame

    >>> midx = pd.MultiIndex(levels=[['lama', 'cow', 'falcon'],
    ...                              ['speed', 'weight', 'length']],
    ...                      codes=[[0, 0, 0, 1, 1, 1, 2, 2, 2],
    ...                             [0, 1, 2, 0, 1, 2, 0, 1, 2]])
    >>> df = pd.DataFrame(index=midx, columns=['big', 'small'],
    ...                   data=[[45, 30], [200, 100], [1.5, 1], [30, 20],
    ...                         [250, 150], [1.5, 0.8], [320, 250],
    ...                         [1, 0.8], [0.3, 0.2]])
    >>> df
                    big     small
    lama    speed   45.0    30.0
            weight  200.0   100.0
            length  1.5     1.0
    cow     speed   30.0    20.0
            weight  250.0   150.0
            length  1.5     0.8
    falcon  speed   320.0   250.0
            weight  1.0     0.8
            length  0.3     0.2

    >>> tf = DropTF(index='cow', columns='small')
    >>> tf.transform(df)
    lama    speed      45.0
            weight    200.0
            length      1.5
    falcon  speed     320.0
            weight      1.0
            length      0.3
    Name: big, dtype: float64

    >>> tf = DropTF(index='length', level=1)
    >>> tf.transform(df)
                    big     small
    lama    speed   45.0    30.0
            weight  200.0   100.0
    cow     speed   30.0    20.0
            weight  250.0   150.0
    falcon  speed   320.0   250.0
            weight  1.0     0.8
    """

    transform_method = staticmethod(pd.DataFrame.drop)



from __future__ import annotations

import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc
from shapely.geometry.base import BaseGeometry

from dtoolkit.geoaccessor.geoseries import geodistance as s_geodistance
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_geodistance)
def geodistance(
    df: gpd.GeoDataFrame,
    /,
    other: BaseGeometry | gpd.GeoSeries | gpd.GeoDataFrame,
    align: bool = True,
    radius: float = 6371008.7714150598,
) -> pd.Series:

    return s_geodistance(df.geometry, other=other, align=align, radius=radius)



from dtoolkit.transformer.pandas.AppendTF import AppendTF  # noqa: F401
from dtoolkit.transformer.pandas.AssignTF import AssignTF  # noqa: F401
from dtoolkit.transformer.pandas.DropTF import DropTF  # noqa: F401
from dtoolkit.transformer.pandas.EvalTF import EvalTF  # noqa: F401
from dtoolkit.transformer.pandas.FillnaTF import FillnaTF  # noqa: F401
from dtoolkit.transformer.pandas.FilterInTF import FilterInTF  # noqa: F401
from dtoolkit.transformer.pandas.FilterTF import FilterTF  # noqa: F401
from dtoolkit.transformer.pandas.GetTF import GetTF  # noqa: F401
from dtoolkit.transformer.pandas.QueryTF import QueryTF  # noqa: F401
from dtoolkit.transformer.pandas.ReplaceTF import ReplaceTF  # noqa: F401
from dtoolkit.transformer.pandas.SelectDtypesTF import SelectDtypesTF  # noqa: F401



import pandas as pd

from dtoolkit.transformer import DataFrameTF


class FilterTF(DataFrameTF):
    """
    A transformer could get subset the dataframe rows or columns according to
    the specified index labels.

    See Also
    --------
    pandas.DataFrame.filter : This transformer's prototype method.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dtoolkit.transformer import FilterTF
    >>> df = pd.DataFrame(np.array(([1, 2, 3], [4, 5, 6])),
    ...                   index=['mouse', 'rabbit'],
    ...                   columns=['one', 'two', 'three'])
    >>> df
            one  two  three
    mouse     1    2      3
    rabbit    4    5      6

    Select columns by name

    >>> tf = FilterTF(items=['one', 'three'])
    >>> tf.transform(df)
                one  three
    mouse     1      3
    rabbit    4      6

    Select columns by regular expression

    >>> tf = FilterTF(regex='e$', axis=1)
    >>> tf.transform(df)
                one  three
    mouse     1      3
    rabbit    4      6

    Select rows containing 'bbi'

    >>> tf = FilterTF(like='bbi', axis=0)
    >>> tf.transform(df)
                one  two  three
    rabbit    4    5      6
    """

    transform_method = staticmethod(pd.DataFrame.filter)



from __future__ import annotations

from typing import Literal

import pandas as pd

from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def boolean(
    df: pd.DataFrame,
    /,
    how: Literal["any", "all"] = "any",
    complement: bool = False,
    **kwargs,
) -> pd.Series:
    """
    Return whether any (or all) element is True, potentially over an ``axis``.

    An API to gather :meth:`~pandas.DataFrame.any` and
    :meth:`~pandas.DataFrame.all` to one.

    Parameters
    ----------
    how : {'any', 'all'}, default 'any'
        Choose a method to get mask.

        * 'any' : Do :meth:`~pandas.DataFrame.any` function to DataFrame.
        * 'all' : Do :meth:`~pandas.DataFrame.all` function to DataFrame.

    axis : {0 or 'index', 1 or 'columns', None}, default 0
        Indicate which axis or axes should be reduced.

        * 0 / 'index' : reduce the index, return a Series whose index is
          the original column labels.

        * 1 / 'columns' : reduce the columns, return a Series whose index is
          the original index.

        * None : reduce all axes, return a scalar.

    complement : bool, default False
        If True does logical 'not' operator to values firstly.

    **kwargs
        See the documentation for :meth:`~pandas.DataFrame.any` and
        :attr:`~pandas.DataFrame.all` for complete details on the keyword arguments.

    Returns
    -------
    Series

    Raises
    ------
    ValueError
        If ``how`` isn't "any" or "all".

    See Also
    --------
    pandas.DataFrame.any
    pandas.DataFrame.all

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({"a": [True, True, False], "b": [False, True, False]})
    >>> df
           a      b
    0   True  False
    1   True   True
    2  False  False

    Get the bool value from each column.

    >>> df.boolean(how="any")
    a    True
    b    True
    dtype: bool

    Get the bool value from each row. And require each element is True.

    >>> df.boolean(how="all", axis=1)
    0    False
    1     True
    2    False
    dtype: bool

    Do a '~' (logical not) operation then get the bool value.

    >>> df.boolean(how="all", axis=1, complement=True)
    0    False
    1    False
    2     True
    dtype: bool

    Get the bool value from the entire data.

    >>> df.boolean(how="all", axis=None)
    False
    """

    if how not in {"any", "all"}:
        raise ValueError(f"invalid how option: {how!r}")

    return getattr(~df if complement else df, how)(**kwargs)



import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def to_datetime(s: pd.Series, /, **kwargs) -> pd.Series:
    """
    Convert Series to datetime type.

    A sugary syntax wraps::

        pd.to_datetime(s, **kwargs)

    Parameters
    ----------
    **kwargs
        Keyword arguments to pass to :meth:`pandas.to_datetime`.

    Returns
    -------
    Series

    See Also
    --------
    pandas.to_datetime

    Notes
    -----
    ``DataFrame.to_datetime`` could use `janitor.functions.to_datetime`__.

    __ https://pyjanitor-devs.github.io/pyjanitor/api/functions/\
#janitor.functions.to_datetime

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series(['20200101', '20200202', '20200303'])
    >>> s
    0    20200101
    1    20200202
    2    20200303
    dtype: object
    >>> s.to_datetime(format='%Y%m%d')
    0   2020-01-01
    1   2020-02-02
    2   2020-03-03
    dtype: datetime64[ns]
    """

    return pd.to_datetime(s, **kwargs)



from functools import wraps
from warnings import warn

from dtoolkit.util._exception import find_stack_level


# based on https://stackoverflow.com/a/49802489/9155078
def deprecated_alias(
    warning_msg: str = (
        "'{func_name}'s paramerter '{old_alias}' is deprecated, "
        "please use the parameter '{new_alias}'."
    ),
    error_msg: str = (
        "'{func_name}' shouldn't received both '{old_alias}' "
        "and '{new_alias}' parameters at the same time."
    ),
    **aliases,
):
    """
    Used as a decorator when deprecating old function argument names, while keeping
    backwards compatibility.

    Parameters
    -----------
    warning_msg : str
        The message of :exc:`DeprecationWarning`. It should be a string or a
        string template. If a string template defaults input ``func_name``,
        ``old_alias`` and ``new_alias``.

    error_msg : str
        The message of :exc:`TypeError`. It should be a string or a string
        template. If a string template defaults input ``func_name``, ``old_alias`` and
        ``new_alias``.

    **aliases : dict
        Dictionary of aliases for a function's arguments,
        like ``{old_alias: new_alias}``.

    Raises
    ------
    DeprecationWarning
        If ``old_alias`` is input into the function.

    TypeError
        If ``old_alias`` and ``new_alias`` are both input into the function.

    See Also
    --------
    dtoolkit.util._decorator.warning
    dtoolkit.util._decorator.deprecated_kwargs

    Examples
    --------
    >>> from dtoolkit.util._decorator import deprecated_alias
    >>> @deprecated_alias(a='alpha', b='beta')
    ... def plus(alpha, beta):
    ...     return alpha + beta
    >>> plus(a=1, beta=2)
    3
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for old_alias, new_alias in aliases.items():
                if old_alias in kwargs:
                    if new_alias in kwargs:
                        raise TypeError(
                            error_msg.format(
                                func_name=func.__name__,
                                old_alias=old_alias,
                                new_alias=new_alias,
                            ),
                        )

                    warn(
                        warning_msg.format(
                            func_name=func.__name__,
                            old_alias=old_alias,
                            new_alias=new_alias,
                        ),
                        DeprecationWarning,
                        find_stack_level(),
                    )

                    kwargs[new_alias] = kwargs.pop(old_alias)

            return func(*args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
import pandas as pd

from dtoolkit.accessor.register import register_series_method

if TYPE_CHECKING:
    from pyproj import CRS


@register_series_method
def from_wkt(
    s: pd.Series,
    /,
    crs: CRS | str | int = None,
    drop: bool = False,
) -> gpd.GeoSeries | gpd.GeoDataFrame:
    """
    Generate :obj:`~geopandas.GeoDataFrame` of geometries from :obj:`~pandas.Series`.

    A sugary syntax wraps :meth:`geopandas.GeoSeries.from_wkt`.

    Parameters
    ----------
    crs : CRS, str, int, optional
        Coordinate Reference System of the geometry objects. Can be anything
        accepted by :meth:`~pyproj.crs.CRS.from_user_input`, such as an authority
        string (eg "EPSG:4326" / 4326) or a WKT string.

    drop : bool, default False
        Don't contain original WKT anymore.

    Returns
    -------
    GeoSeries or GeoDataFrame
        GeoSeries if `drop` is True else GeoDataFrame.

    Raises
    ------
    ValueError
        If ``drop`` is False and the name of Series is empty.

    See Also
    --------
    geopandas.GeoSeries.from_wkt
    dtoolkit.geoaccessor.dataframe.from_wkt

    Notes
    -----
    This method is the accessor of Series, not GeoSeries.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> s = pd.Series(
    ...     [
    ...         "POINT (1 1)",
    ...         "POINT (2 2)",
    ...         "POINT (3 3)",
    ...     ],
    ...     name='wkt',
    ... )
    >>> s
    0    POINT (1 1)
    1    POINT (2 2)
    2    POINT (3 3)
    Name: wkt, dtype: object
    >>> s.from_wkt(crs=4326)
               wkt                 geometry
    0  POINT (1 1)  POINT (1.00000 1.00000)
    1  POINT (2 2)  POINT (2.00000 2.00000)
    2  POINT (3 3)  POINT (3.00000 3.00000)

    Drop original 'wkt'.

    >>> s.from_wkt("wkt", drop=True)
    0    POINT (1.000 1.000)
    1    POINT (2.000 2.000)
    2    POINT (3.000 3.000)
    dtype: geometry
    """

    if not drop and s.name is None:
        raise ValueError(
            "to keep the original data requires setting the 'name' of "
            f"{s.__class__.__name__!r}",
        )

    geometry = gpd.GeoSeries.from_wkt(s, crs=crs)
    return geometry if drop else gpd.GeoDataFrame(s, geometry=geometry)



from __future__ import annotations

from typing import Literal

import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geodataframe.duplicated_geometry import duplicated_geometry
from dtoolkit.geoaccessor.geoseries import (
    drop_duplicates_geometry as s_drop_duplicates_geometry,
)
from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import BINARY_PREDICATE
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_drop_duplicates_geometry, klass="GeoDataFrame")
def drop_duplicates_geometry(
    df: gpd.GeoDataFrame,
    /,
    predicate: BINARY_PREDICATE | None = None,
    keep: Literal["first", "last", False] = "first",
) -> gpd.GeoDataFrame:
    return df[~duplicated_geometry(df, predicate=predicate, keep=keep)]



from __future__ import annotations

from typing import Hashable
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from dtoolkit.accessor.dataframe.drop_or_not import drop_or_not
from dtoolkit.accessor.register import register_dataframe_method
from dtoolkit.accessor.series.expand import collapse

if TYPE_CHECKING:
    from sklearn.base import TransformerMixin


@register_dataframe_method
def decompose(
    df: pd.DataFrame,
    /,
    method: TransformerMixin,
    columns: None
    | dict[Hashable | tuple[Hashable], Hashable | list[Hashable] | tuple[Hashable]]
    | list[Hashable]
    | pd.Index = None,
    drop: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """
    Decompose DataFrame's columns.

    Parameters
    ----------
    method : TransformerMixin
        Decomposition transformer.

    columns : dict, Series, list, tuple or None, default None
        Choose columns to decompose.

        - None : Decompose all columns.
        - list or Index : Decompose the selected columns.
        - dict : Decompose and remap columns to a few, ``{new columns: old columns}``.

    drop : bool, default False
        If True, drop the used columns when ``columns`` is :keyword:`dict`.

    **kwargs
        See the documentation for ``method`` for complete details on
        the keyword arguments.

    Returns
    -------
    DataFrame

    Raises
    ------
    ValueError
        If the number of rows is less than the number of columns.

    See Also
    --------
    sklearn.decomposition
        Scikit-learn's matrix decomposition transformer.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> from sklearn import decomposition
    >>> df = pd.DataFrame(
    ...     [
    ...         [-1, -1, 1, 1],
    ...         [-2, -1, 2, 1],
    ...         [-3, -2, 3, 2],
    ...         [1, 1, -1, -1],
    ...         [2, 1, -2, -1],
    ...         [3, 2, -3, -2],
    ...     ],
    ...     columns=["a", "b", "c", "d"],
    ... )
    >>> df
       a  b  c  d
    0 -1 -1  1  1
    1 -2 -1  2  1
    2 -3 -2  3  2
    3  1  1 -1 -1
    4  2  1 -2 -1
    5  3  2 -3 -2

    Decompose all columns.

    >>> df.decompose(decomposition.PCA)  # doctest: +SKIP
              a         b             c             d
    0  1.956431  0.415183  9.009015e-17  8.100537e-18
    1  3.142238 -0.355441  8.394617e-17  9.817066e-18
    2  5.098670  0.059742 -8.445140e-17  1.640353e-19
    3 -1.956431 -0.415183 -7.881266e-17  8.428608e-18
    4 -3.142238  0.355441 -8.495664e-17  1.014514e-17
    5 -5.098670 -0.059742  8.445140e-17 -1.640353e-19

    Decompose the selected columns.

    >>> df.decompose(decomposition.PCA, ["a", "b"])  # doctest: +SKIP
              a         b  c  d
    0  1.383406  0.293579  1  1
    1  2.221898 -0.251335  2  1
    2  3.605304  0.042244  3  2
    3 -1.383406 -0.293579 -1 -1
    4 -2.221898  0.251335 -2 -1
    5 -3.605304 -0.042244 -3 -2

    Decompose and remap columns to a few.

    >>> df.decompose(
    ...     decomposition.PCA,
    ...     {"A": ["a", "b"], "B": ["b", "c", "d"]},
    ... )  # doctest: +SKIP
              A         B  a  b  c  d
    0  1.383406  1.694316 -1 -1  1  1
    1  2.221898  2.428593 -2 -1  2  1
    2  3.605304  4.122909 -3 -2  3  2
    3 -1.383406 -1.694316  1  1 -1 -1
    4 -2.221898 -2.428593  2  1 -2 -1
    5 -3.605304 -4.122909  3  2 -3 -2
    >>> df.decompose(
    ...     decomposition.PCA,
    ...     {("A", "B"): ["a", "b", "c"]}
    ... )  # doctest: +SKIP
              A         B  a  b  c  d
    0  1.702037  0.321045 -1 -1  1  1
    1  2.988071 -0.267273 -2 -1  2  1
    2  4.690108  0.053773 -3 -2  3  2
    3 -1.702037 -0.321045  1  1 -1 -1
    4 -2.988071  0.267273  2  1 -2 -1
    5 -4.690108 -0.053773  3  2 -3 -2
    """

    if columns is None:
        return pd.DataFrame(
            _decompose(df, method, **kwargs),
            index=df.index,
            columns=df.columns,
        )

    elif isinstance(columns, (list, pd.Index)):
        return pd.DataFrame(
            _decompose(df[columns], method, **kwargs),
            index=df.index,
            columns=columns,
        ).combine_first(df)

    elif isinstance(columns, dict):
        return pd.DataFrame(
            np.hstack(
                [
                    _decompose(
                        df[value],
                        method,
                        n_components=len(key) if isinstance(key, tuple) else 1,
                        **kwargs,
                    )
                    for key, value in columns.items()
                ],
            ),
            index=df.index,
            columns=collapse(columns.keys()),
        ).combine_first(
            drop_or_not(
                df,
                drop=drop,
                columns=collapse(columns.values()),
            ),
        )

    raise ValueError("The type of inputting 'columns' isn't right")


def _decompose(
    df: pd.DataFrame,
    /,
    method: TransformerMixin,
    n_components=None,
    **kwargs,
) -> np.ndarray:
    if n_components is None and len(df) < df.columns.size:
        raise ValueError(
            "Don't support decomposing DataFrame in which "
            "the number of rows is less than the number of columns",
        )

    return method(n_components, **kwargs).fit_transform(
        df.to_frame() if isinstance(df, pd.Series) else df,
    )



from dtoolkit.geoaccessor.geoseries.cncrs_offset import cncrs_offset  # noqa: F401
from dtoolkit.geoaccessor.geoseries.count_coordinates import (  # noqa: F401
    count_coordinates,
)
from dtoolkit.geoaccessor.geoseries.drop_duplicates_geometry import (  # noqa: F401
    drop_duplicates_geometry,
)
from dtoolkit.geoaccessor.geoseries.duplicated_geometry import (  # noqa: F401
    duplicated_geometry,
)
from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import (  # noqa: F401
    duplicated_geometry_groups,
)
from dtoolkit.geoaccessor.geoseries.filter_geometry import filter_geometry  # noqa: F401
from dtoolkit.geoaccessor.geoseries.geoarea import geoarea  # noqa: F401
from dtoolkit.geoaccessor.geoseries.geobuffer import geobuffer  # noqa: F401
from dtoolkit.geoaccessor.geoseries.geodistance import geodistance  # noqa: F401
from dtoolkit.geoaccessor.geoseries.geodistance_matrix import (  # noqa: F401
    geodistance_matrix,
)
from dtoolkit.geoaccessor.geoseries.get_coordinates import get_coordinates  # noqa: F401
from dtoolkit.geoaccessor.geoseries.has_hole import has_hole  # noqa: F401
from dtoolkit.geoaccessor.geoseries.hole_counts import hole_counts  # noqa: F401
from dtoolkit.geoaccessor.geoseries.reverse_geocode import reverse_geocode  # noqa: F401
from dtoolkit.geoaccessor.geoseries.toposimplify import toposimplify  # noqa: F401
from dtoolkit.geoaccessor.geoseries.xy import xy  # noqa: F401



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(alias="s")
def geoarea(s: gpd.GeoSeries, /) -> pd.Series:
    r"""
    Returns a ``Series`` containing the **geographic area** (m\ :sup:`2`) of each
    geometry.

    A sugar syntax wraps::

        {alias}.to_crs("+proj=cea").area

    Returns
    -------
    Series

    See Also
    --------
    geopandas.GeoSeries.area
    dtoolkit.geoaccessor.geoseries.geoarea
    dtoolkit.geoaccessor.geodataframe.geoarea

    Notes
    -----
    The result is a tiny bit different from the value, because of CRS problem. But the
    `cea`_ (Equal Area Cylindrical) CRS is quite enough, the average absolute error is
    less than 0.04% base on 'naturalearth_lowres' data.

    .. _cea: https://proj.org/operations/projections/cea.html

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> df = gpd.read_file(gpd.datasets.get_path("naturalearth_lowres"))
    >>> df.head()
           pop_est  ...                                           geometry
    0     889953.0  ...  MULTIPOLYGON (((180.00000 -16.06713, 180.00000...
    1   58005463.0  ...  POLYGON ((33.90371 -0.95000, 34.07262 -1.05982...
    2     603253.0  ...  POLYGON ((-8.66559 27.65643, -8.66512 27.58948...
    3   37589262.0  ...  MULTIPOLYGON (((-122.84000 49.00000, -122.9742...
    4  328239523.0  ...  MULTIPOLYGON (((-122.84000 49.00000, -120.0000...
    <BLANKLINE>
    [5 rows x 6 columns]
    >>> df.geoarea().head()
    0    1.928760e+10
    1    9.327793e+11
    2    9.666925e+10
    3    1.003773e+13
    4    9.509851e+12
    dtype: float64
    """

    return s.to_crs("+proj=cea").area



from __future__ import annotations

from typing import Hashable
from typing import TYPE_CHECKING

import geopandas as gpd
import pandas as pd

from dtoolkit.accessor.register import register_dataframe_method

if TYPE_CHECKING:
    from pyproj import CRS


@register_dataframe_method
def to_geoframe(
    df: pd.DataFrame,
    /,
    geometry: Hashable | gpd.GeoSeries = None,
    crs: CRS | str | int = None,
    **kwargs,
) -> gpd.GeoDataFrame:
    """
    Transform a :class:`~pandas.DataFrame` to a :class:`~geopandas.GeoDataFrame`.

    Parameters
    ----------
    geometry : Hashable or GeoSeries, optional
        If str or int, column to use as geometry. If array, will be set as 'geometry'
        column on GeoDataFrame.

    crs : CRS, str, int, optional
        Coordinate Reference System of the geometry objects. Can be anything
        accepted by :meth:`~pyproj.crs.CRS.from_user_input`, such as an authority
        string (eg "EPSG:4326" / 4326) or a WKT string.

    **kwargs
        See the documentation for :class:`~geopandas.GeoDataFrame` and  for complete
        details on the keyword arguments.

    Returns
    -------
    GeoDataFrame

    See Also
    --------
    dtoolkit.geoaccessor.series.to_geoseries
    dtoolkit.geoaccessor.series.to_geoframe

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df_point = (
    ...     pd.DataFrame({"x": [122, 100], "y":[55, 1]})
    ...     .from_xy("x", "y", drop=True, crs=4326)
    ... )
    >>> df_point
                         geometry
    0  POINT (122.00000 55.00000)
    1   POINT (100.00000 1.00000)
    >>> df_data = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    >>> df = pd.concat((df_data, df_point), axis=1)
    >>> df
       a  b                    geometry
    0  1  3  POINT (122.00000 55.00000)
    1  2  4   POINT (100.00000 1.00000)

    ``df`` is DataFrame type not GeoDataFrame type.

    >>> type(df)
    <class 'pandas.core.frame.DataFrame'>

    >>> df = df.to_geoframe()
    >>> df
       a  b                    geometry
    0  1  3  POINT (122.00000 55.00000)
    1  2  4   POINT (100.00000 1.00000)

    ``df`` is GeoDataFrame type at present.

    >>> type(df)
    <class 'geopandas.geodataframe.GeoDataFrame'>

    Use original CRS if its CRS is not specified.

    >>> df.crs == df_point.crs == 4326
    True
    >>> df.crs
    <Geographic 2D CRS: EPSG:4326>
    Name: WGS 84
    Axis Info [ellipsoidal]:
    - Lat[north]: Geodetic latitude (degree)
    - Lon[east]: Geodetic longitude (degree)
    Area of Use:
    - name: World.
    - bounds: (-180.0, -90.0, 180.0, 90.0)
    Datum: World Geodetic System 1984 ensemble
    - Ellipsoid: WGS 84
    - Prime Meridian: Greenwich
    """

    return (
        gpd.GeoDataFrame(
            # Avoid mutating the original DataFrame.
            # https://github.com/geopandas/geopandas/issues/1179
            df.copy(),
            crs=crs,
            geometry=geometry,
            **kwargs,
        )
        if not isinstance(df, gpd.GeoDataFrame)
        else df
    )



import pandas as pd

from dtoolkit.transformer.base import DataFrameTF


class SelectDtypesTF(DataFrameTF):
    """
    A transformer could return a subset of the :obj:`~pandas.DataFrame`'s
    columns based on the column dtypes.

    See Also
    --------
    pandas.DataFrame.select_dtypes : This transformer's prototype method.

    Examples
    --------
    >>> import pandas as pd
    >>> from dtoolkit.transformer import SelectDtypesTF
    >>> df = pd.DataFrame({'a': [1, 2] * 3,
    ...                    'b': [True, False] * 3,
    ...                    'c': [1.0, 2.0] * 3})
    >>> df
            a      b  c
    0       1   True  1.0
    1       2  False  2.0
    2       1   True  1.0
    3       2  False  2.0
    4       1   True  1.0
    5       2  False  2.0

    >>> tf = SelectDtypesTF(include='bool')
    >>> tf.transform(df)
    0     True
    1    False
    2     True
    3    False
    4     True
    5    False
    Name: b, dtype: bool

    >>> tf = SelectDtypesTF(include=['float64'])
    >>> tf.transform(df)
    0    1.0
    1    2.0
    2    1.0
    3    2.0
    4    1.0
    5    2.0
    Name: c, dtype: float64

    >>> tf = SelectDtypesTF(exclude=['int64'])
    >>> tf.transform(df)
            b    c
    0   True  1.0
    1  False  2.0
    2   True  1.0
    3  False  2.0
    4   True  1.0
    5  False  2.0
    """

    transform_method = staticmethod(pd.DataFrame.select_dtypes)



import inspect
import os.path

import dtoolkit


# based on pandas/util/_exceptions.py
def find_stack_level() -> int:
    """
    Find the first place in the stack that is not inside dtoolkit
    (test notwithstanding).

    Returns
    -------
    int
        The first place in the stack level.

    Examples
    --------
    >>> from warnings import warn
    >>> from dtoolkit.util._exception import find_stack_level
    >>> warn("Blah blah blah", stacklevel=find_stack_level())
    """

    pkg_dir = os.path.abspath(os.path.dirname(dtoolkit.__file__))
    test_dir = os.path.join(os.path.dirname(pkg_dir), "test")

    # https://stackoverflow.com/questions/17407119/python-inspect-stack-is-slow
    n = 0
    frame = inspect.currentframe()
    while frame:
        fname = inspect.getfile(frame)
        if not fname.startswith(pkg_dir) or fname.startswith(test_dir):
            break

        frame = frame.f_back
        n += 1

    return n



from __future__ import annotations

from warnings import catch_warnings
from warnings import simplefilter

import geopandas as gpd
import numpy as np
import pandas as pd
from pandas.api.types import is_list_like
from pandas.api.types import is_number
from pandas.util._decorators import doc

from dtoolkit._typing import Number
from dtoolkit._typing import OneDimArray
from dtoolkit.geoaccessor.geoseries.xy import xy  # noqa: F401
from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(klass="GeoSeries", alias="s")
def geobuffer(
    s: gpd.GeoSeries,
    distance: Number | list[Number] | OneDimArray,
    /,
    **kwargs,
) -> gpd.GeoSeries:
    """
    Creates geographic buffers for :class:`~geopandas.{klass}`.

    Reprojects input features into the *UTM* projection, buffers them,
    then reprojects back into the original geographic coordinates.

    Parameters
    ----------
    distance : int, float, list-like of int or float, the unit is meter.
        The radius of the buffer. If numpy.ndarray or Series are used then it must have
        same length as the ``{alias}``. For ``GeoDataFrame.geobuffer``, it would use the
        column name as the distance prior.

    Returns
    -------
    {klass}

    Warns
    -----
    UserWarning
        - If the index of the inputting is not unique.
        - If the CRS of the inputting is not WGS84 (epsg:4326).

    Raises
    ------
    ValueError
        Requires the CRS of the inputting is WGS84 (epsg:4326).

    TypeError
        If ``distance`` is not a number.

    IndexError
        - If ``distance`` is a list-like but its length does not match the length of
          ``{alias}``.
        - If ``distance`` is a Series but its index does not match the index of
          ``{alias}``.

    See Also
    --------
    geopandas.GeoSeries.buffer
    dtoolkit.geoaccessor.geoseries.geobuffer
    dtoolkit.geoaccessor.geodataframe.geobuffer

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df = (
    ...     pd.DataFrame(
    ...         {{
    ...             "distance": [0, 10],
    ...             "where": ["close to equator", "away from equator"],
    ...             "x": [122, 100],
    ...             "y": [55, 1],
    ...         }},
    ...     )
    ...     .from_xy(
    ...         "x",
    ...         "y",
    ...         crs=4326,
    ...         drop=True,
    ...     )
    ... )
    >>> df
       distance              where                    geometry
    0         0   close to equator  POINT (122.00000 55.00000)
    1        10  away from equator   POINT (100.00000 1.00000)
    >>> df.geobuffer(100)
       distance  ...                                           geometry
    0         0  ...  POLYGON ((122.00156 55.00001, 122.00156 54.999...
    1        10  ...  POLYGON ((100.00090 1.00000, 100.00089 0.99991...
    <BLANKLINE>
    [2 rows x 3 columns]

    For GeoDataFrame, it could use the column name as the distance to generate buffer.

    >>> df.geobuffer("distance")
       distance  ...                                           geometry
    0         0  ...                                      POLYGON EMPTY
    1        10  ...  POLYGON ((100.00009 1.00000, 100.00009 0.99999...
    <BLANKLINE>
    [2 rows x 3 columns]
    """

    if s.crs != 4326:
        raise ValueError(
            f"The CRS is {s.crs}, which requires is 'WGS86' (EPSG:4326).",
        )

    distance_is_list = is_list_like(distance)
    if not (is_number(distance) or distance_is_list):
        raise TypeError(
            "'distance' should be a number or a list of number, "
            f"but got {type(distance)!r}.",
        )

    if isinstance(distance, pd.Series) and not s.index.equals(distance.index):
        raise IndexError(
            "Index values of 'distance' sequence doesn't "
            f"match index values of the {type(s)!r}",
        )

    if distance_is_list:
        distance = np.asarray(distance)
        if distance.size != s.size:
            raise IndexError(
                f"Length of 'distance' doesn't match length of the {type(s)!r}.",
            )

    with catch_warnings():
        # Ignore UserWarning ("Geometry is in a geographic CRS")
        simplefilter("ignore", UserWarning)
        utms = s.centroid.xy().apply(wgs_to_utm).to_numpy()

    s = s.copy()
    for utm in pd.unique(utms):
        if utm is None:
            continue

        mask = utms == utm
        dis = distance[mask] if distance_is_list else distance
        s[mask] = s[mask].to_crs(utm).buffer(dis, **kwargs).to_crs(4326)

    return s


def wgs_to_utm(point: tuple[float, float]) -> str | None:
    """Based on (x, y), return the best UTM EPSG code."""

    x, y = point
    if is_number(x) and -180 <= x <= 180 and is_number(y) and -90 <= y <= 90:
        zone = (x + 180) // 6 % 60 + 1
        return f"EPSG:{326 if y >= 0 else 327}{zone:02.0f}"



from __future__ import annotations

import geopandas as gpd
import numpy as np
import pandas as pd

from dtoolkit.geoaccessor.geoseries.xy import xy
from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
def geodistance_matrix(
    s: gpd.GeoSeries,
    /,
    other: gpd.GeoSeries | gpd.GeoDataFrame | None = None,
    radius: float = 6371008.7714150598,
) -> pd.DataFrame:
    """
    Returns a ``DataFrame`` containing the `great-circle`__ distances matrix between in
    ``s`` and ``other`` via haversine formula.

    __ https://en.wikipedia.org/wiki/Great-circle_distance

    .. math::

        D(x, y) = 2 \\arcsin [
            \\sqrt{
                \\sin^2 ((y_2 - y_1) / 2)
                + \\cos(y_1) \\cos(y_1) \\sin^2 ((x_2 - y_1) / 2)
            }
        ]

    Parameters
    ----------
    other : GeoSeries, or GeoDataFrame, default None
        If None, uses ``other=s``.

    radius : float, default 6371008.7714150598
        Great-circle distance uses a spherical model of the earth, using the mean earth
        radius as defined by the International Union of Geodesy and Geophysics,
        (2\\ *a* + *b*)/3 = 6371008.7714150598 meters for WGS-84.

    Returns
    -------
    DataFrame
        - The index and columns are the same as the index of ``s`` and ``other``.
        - The values are the great-circle distances and its unit is meters.

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'sklearn'.

    ValueError
        If the CRS is not ``ESGP:4326``.

    TypeError
        If the other is not a GeoSeries, GeoDataFrame, or None type.

    See Also
    --------
    sklearn.metrics.pairwise.haversine_distances
    dtoolkit.geoaccessor.geoseries.geodistance
    dtoolkit.geoaccessor.geoseries.geodistance_matrix
    dtoolkit.geoaccessor.geodataframe.geodistance
    dtoolkit.geoaccessor.geodataframe.geodistance_matrix

    Notes
    -----
    - Currently, only supports Point geometry.
    - The great-circle distance is the angular distance between two points on the
      surface of a sphere. As the Earth is nearly spherical, the haversine formula
      provides a good approximation of the distance between two points of the Earth
      surface, with a less than 1% error on average.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     {
    ...         "x": [120, 122, 100],
    ...         "y":[30, 55, 1],
    ...     },
    ... ).from_xy("x", "y", crs=4326)
    >>> df
         x   y                    geometry
    0  120  30  POINT (120.00000 30.00000)
    1  122  55  POINT (122.00000 55.00000)
    2  100   1   POINT (100.00000 1.00000)
    >>> other = pd.DataFrame(
    ...     {
    ...         "x": [120, 110],
    ...         "y":[30, 40],
    ...     },
    ... ).from_xy("x", "y", crs=4326)
    >>> other
         x   y                    geometry
    0  120  30  POINT (120.00000 30.00000)
    1  110  40  POINT (110.00000 40.00000)
    >>> df.geodistance_matrix(other)
                  0             1
    0  0.000000e+00  1.435335e+06
    1  2.784435e+06  1.889892e+06
    2  3.855604e+06  4.453100e+06
    """
    from sklearn.metrics.pairwise import haversine_distances

    if s.crs != 4326:
        raise ValueError(f"Only support 'EPSG:4326' CRS, but got {s.crs!r}.")

    if other is None:
        Y = None
    elif isinstance(other, gpd.base.GeoPandasBase):
        if other.crs != 4326:
            raise ValueError(f"Only support 'EPSG:4326' CRS, but got {other.crs!r}.")

        Y = np.radians(xy(other.geometry, reverse=True).tolist())
    else:
        raise TypeError(f"Unknown type: {type(other).__name__!r}.")

    X = np.radians(xy(s, reverse=True).tolist())
    return pd.DataFrame(
        radius * haversine_distances(X, Y),
        index=s.index,
        columns=other.index if other is not None else s.index,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
import pandas as pd

from dtoolkit.accessor.register import register_series_method

if TYPE_CHECKING:
    from pyproj import CRS


@register_series_method
def from_wkb(
    s: pd.Series,
    /,
    crs: CRS | str | int = None,
    drop: bool = False,
) -> gpd.GeoSeries | gpd.GeoDataFrame:
    """
    Generate :obj:`~geopandas.GeoDataFrame` of geometries from :obj:`~pandas.Series`.

    A sugary syntax wraps :meth:`geopandas.GeoSeries.from_wkb`.

    Parameters
    ----------
    crs : CRS, str, int, optional
        Coordinate Reference System of the geometry objects. Can be anything
        accepted by :meth:`~pyproj.crs.CRS.from_user_input`, such as an authority
        string (eg "EPSG:4326" / 4326) or a WKT string.

    drop : bool, default False
        Don't contain original WKB anymore.

    Returns
    -------
    GeoSeries or GeoDataFrame
        GeoSeries if `drop` is True else GeoDataFrame.

    Raises
    ------
    ValueError
        If ``drop`` is False and the name of Series is empty.

    See Also
    --------
    geopandas.GeoSeries.from_wkb
    dtoolkit.geoaccessor.series.from_wkt
    dtoolkit.geoaccessor.dataframe.from_wkb

    Notes
    -----
    - This method is the accessor of Series, not GeoSeries.
    - Read from file (such as "CSV" or "EXCEL"), requreis converting "WKB" column
      type from ``str`` to ``bytes`` via ``eval``.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> s = pd.Series(
    ...      [
    ...          "POINT (1 1)",
    ...          "POINT (2 2)",
    ...          "POINT (3 3)",
    ...      ],
    ...      name="wkt",
    ... )
    >>> s
    0    POINT (1 1)
    1    POINT (2 2)
    2    POINT (3 3)
    Name: wkt, dtype: object
    >>> s_wkb = s.from_wkt(crs=4326, drop=True).to_wkb().rename('wkb')
    >>> s_wkb  # doctest: +SKIP
    0    b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00...'
    1    b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00...'
    2    b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00...'
    Name: wkb, dtype: object
    >>> type(s_wkb)
    <class 'pandas.core.series.Series'>
    >>> gdf = s_wkb.from_wkb(crs=4326)
    >>> gdf  # doctest: +SKIP
                                                      wkb                 geometry
    0  b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00...'  POINT (1.00000 1.00000)
    1  b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00...'  POINT (2.00000 2.00000)
    2  b'\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00...'  POINT (3.00000 3.00000)
    >>> type(gdf)
    <class 'geopandas.geodataframe.GeoDataFrame'>

    Drop original 'wkb' column.

    >>> gs = s_wkb.from_wkb(crs=4326, drop=True)
    >>> gs
    0    POINT (1.00000 1.00000)
    1    POINT (2.00000 2.00000)
    2    POINT (3.00000 3.00000)
    dtype: geometry
    >>> type(gs)
    <class 'geopandas.geoseries.GeoSeries'>
    """

    if not drop and s.name is None:
        raise ValueError(
            "to keep the original data requires setting the 'name' of "
            f"{s.__class__.__name__!r}",
        )

    geometry = gpd.GeoSeries.from_wkb(s, crs=crs)
    return geometry if drop else gpd.GeoDataFrame(s, geometry=geometry)



from __future__ import annotations

from functools import partial
from typing import Callable
from typing import Iterable
from typing import Literal


def parallelize(
    func: Callable,
    jobs: Iterable,
    *,
    n_jobs: int = -1,
    verbose: int = 0,
    timeout: float = None,
    backend: Literal["loky", "multiprocessing", "threading"] = "loky",
    require: Literal["sharedmem"] = None,
    mmap_mode: Literal[None, "r+", "r", "w+", "c"] = "r",
    **kwargs,
):
    """
    Parallelize ``func`` to do ``jobs``.

    A sugary syntax wraps :class:`joblib.Parallel`::

        from functools import partial

        from joblib import Parallel, delayed

        func = delayed(partial(func, **kwargs))
        Parallel(**parallel_kwargs)(map(func, jobs))

    Parameters
    ----------
    func : Callable

    jobs : Iterable

    n_jobs, verbose, timeout, backend, require, mmap_mode
        See the documentation for :class:`joblib.Parallel` for complete details on
        the keyword arguments.

    **kwargs
        See the documentation for ``func`` for complete details on the arguments.

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'joblib'.

    See Also
    --------
    joblib.Parallel
    joblib.delayed

    Examples
    --------
    >>> from dtoolkit.util import parallelize
    >>> parallelize(lambda x: x ** 2, range(3))
    [0, 1, 4]
    """
    from joblib import Parallel, delayed

    func = delayed(partial(func, **kwargs))
    return Parallel(
        n_jobs=n_jobs,
        verbose=verbose,
        timeout=timeout,
        backend=backend,
        require=require,
        mmap_mode=mmap_mode,
    )(map(func, jobs))



from __future__ import annotations

from typing import Literal

import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries.duplicated_geometry import duplicated_geometry
from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import BINARY_PREDICATE
from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(klass="GeoSeries")
def drop_duplicates_geometry(
    s: gpd.GeoSeries,
    /,
    predicate: BINARY_PREDICATE | None = None,
    keep: Literal["first", "last", False] = "first",
) -> gpd.GeoSeries:
    """
    Remove duplicate geometry rows.

    Parameters
    ----------
    predicate : {{'intersects', 'crosses', 'overlaps', 'touches', 'covered_by', \
'contains_properly', 'contains', 'within', 'covers'}}, default None
        The binary predicate is used to validate whether the geometries are duplicates
        or not. If None, the geometries will directly compares via value relation
        instead of the spatial relation.

        .. versionchanged:: 0.0.19
            The 'predicate' default value is changed from 'intersects' to None.

    keep : {{'first', 'last', False}}, default 'first'
        - ``first`` : Mark duplicates as ``True`` except for the first occurrence.
        - ``last`` : Mark duplicates as ``True`` except for the last occurrence.
        - False : Mark all duplicates as ``True``.

    Returns
    -------
    {klass}

    See Also
    --------
    geopandas.sjoin
    dtoolkit.geoaccessor.geoseries.duplicated_geometry
    dtoolkit.geoaccessor.geoseries.drop_duplicates_geometry
    dtoolkit.geoaccessor.geodataframe.duplicated_geometry
    dtoolkit.geoaccessor.geodataframe.drop_duplicates_geometry

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> from shapely.geometry import Polygon
    >>> df = gpd.GeoDataFrame(
    ...     geometry=[
    ...         Polygon([(0,0), (1,0), (1,1), (0,1)]),
    ...         Polygon([(1,1), (2,1), (2,2), (1,2)]),
    ...         Polygon([(2,2), (3,2), (3,3), (2,3)]),
    ...         Polygon([(2, 0), (3, 0), (3, 1)]),
    ...     ],
    ... )
    >>> df
                                                geometry
    0  POLYGON ((0.00000 0.00000, 1.00000 0.00000, 1....
    1  POLYGON ((1.00000 1.00000, 2.00000 1.00000, 2....
    2  POLYGON ((2.00000 2.00000, 3.00000 2.00000, 3....
    3  POLYGON ((2.00000 0.00000, 3.00000 0.00000, 3....

    Work for GeoSeries.

    >>> df.geometry.drop_duplicates_geometry('intersects')
    0    POLYGON ((0.00000 0.00000, 1.00000 0.00000, 1....
    3    POLYGON ((2.00000 0.00000, 3.00000 0.00000, 3....
    Name: geometry, dtype: geometry

    Work for GeoDataFrame too.

    >>> df.drop_duplicates_geometry('intersects')
                                                geometry
    0  POLYGON ((0.00000 0.00000, 1.00000 0.00000, 1....
    3  POLYGON ((2.00000 0.00000, 3.00000 0.00000, 3....
    """

    return s[~duplicated_geometry(s, predicate=predicate, keep=keep)]



from dtoolkit.transformer.base import DataFrameTF  # noqa: F401
from dtoolkit.transformer.base import MethodTF  # noqa: F401
from dtoolkit.transformer.base import NumpyTF  # noqa: F401
from dtoolkit.transformer.base import Transformer  # noqa: F401
from dtoolkit.transformer.factory import methodtf_factory  # noqa: F401
from dtoolkit.transformer.numpy import RavelTF  # noqa: F401
from dtoolkit.transformer.pandas import AppendTF  # noqa: F401
from dtoolkit.transformer.pandas import AssignTF  # noqa: F401
from dtoolkit.transformer.pandas import DropTF  # noqa: F401
from dtoolkit.transformer.pandas import EvalTF  # noqa: F401
from dtoolkit.transformer.pandas import FillnaTF  # noqa: F401
from dtoolkit.transformer.pandas import FilterInTF  # noqa: F401
from dtoolkit.transformer.pandas import FilterTF  # noqa: F401
from dtoolkit.transformer.pandas import GetTF  # noqa: F401
from dtoolkit.transformer.pandas import QueryTF  # noqa: F401
from dtoolkit.transformer.pandas import ReplaceTF  # noqa: F401
from dtoolkit.transformer.pandas import SelectDtypesTF  # noqa: F401
from dtoolkit.transformer.sklearn import OneHotEncoder  # noqa: F401



from dtoolkit.util._decorator.deprecated_alias import deprecated_alias  # noqa: F401
from dtoolkit.util._decorator.deprecated_kwargs import deprecated_kwargs  # noqa: F401
from dtoolkit.util._decorator.warning import warning  # noqa: F401



from __future__ import annotations

import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import geodistance_matrix as s_geodistance_matrix
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_geodistance_matrix)
def geodistance_matrix(
    df: gpd.GeoDataFrame,
    /,
    other: gpd.GeoSeries | gpd.GeoDataFrame | None = None,
    radius: float = 6371008.7714150598,
) -> pd.DataFrame:

    return s_geodistance_matrix(df.geometry, other=other, radius=radius)



from __future__ import annotations

from typing import Hashable

import numpy as np
import pandas as pd
from pandas.api.types import is_number

from dtoolkit._typing import Axis
from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def repeat(
    df: pd.DataFrame,
    repeats: int | Hashable | list[int],
    /,
    axis: Axis = 0,
) -> pd.DataFrame:
    """
    Repeat row or column of a :obj:`~pandas.DataFrame`.

    Returns a new DataFrame where each row/column is repeated
    consecutively a given number of times.

    A sugary syntax wraps :meth:`numpy.repeat`.

    Parameters
    ----------
    repeats : int, Hashable or array of ints
        The number of repetitions for each element. This should be a non-negative
        integer. Repeating 0 times will return an empty :obj:`~pandas.DataFrame`.
        The order of priority type is ``int`` > ``Hashable``.

        * int : the row or column will be repeated ``repeats`` times.

        * array of int : the row or column at the i-th position will be repeated.
          Its length must be the same as the axis being repeated.

        * Hahsable : the row or column will be repeated by with the given column.

    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis along which to repeat.

        * 0, or 'index' : Along the row to repeat.
        * 1, or 'columns' : Along the column to repeat.

    Returns
    -------
    DataFrame
        Newly created DataFrame with repeated elements.

    See Also
    --------
    numpy.repeat : This transformer's prototype method.

    Examples
    --------
    >>> import pandas as pd
    >>> import dtoolkit
    >>> df = pd.DataFrame({'a': [1, 2], 'b':[3, 4]})
    >>> df
       a  b
    0  1  3
    1  2  4

    Each row repeat two times.

    >>> df.repeat(2)
       a  b
    0  1  3
    0  1  3
    1  2  4
    1  2  4

    Each column repeat two times.

    >>> df.repeat(2, 1)
       a  a  b  b
    0  1  1  3  3
    1  2  2  4  4

    ``a`` column repeat 1 times, ``b`` column repeat 2 times.

    >>> df.repeat([1, 2], 1)
       a  b  b
    0  1  3  3
    1  2  4  4

    Use the 'a' column to repeat the row.

    >>> df.repeat('a')
       a  b
    0  1  3
    1  2  4
    1  2  4
    """

    if not is_number(repeats) and isinstance(repeats, Hashable):
        repeats = df[repeats]

    axis = df._get_axis_number(axis)
    return pd.DataFrame(
        np.repeat(df._values, repeats, axis=axis),
        index=df.index.repeat(repeats) if axis == 0 else df.index,
        columns=df.columns.repeat(repeats) if axis == 1 else df.columns,
    )



import pandas as pd

from dtoolkit.transformer import DataFrameTF


class FillnaTF(DataFrameTF):
    """
    A transformer could fill ``NA``/``NaN`` values using the specified method.

    See Also
    --------
    pandas.DataFrame.fillna : This transformer's prototype method.

    Notes
    -----
    :meth:`~pandas.DataFrame.fillna`'s ``inplace`` parameter is not work for
    this transformer. Actually this break pipeline stream. If a transformer's
    ``inplace`` is ``True``, the next tf input would get ``None``.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from dtoolkit.transformer import FillnaTF
    >>> df = pd.DataFrame(
    ...     [
    ...         [np.nan, 2, np.nan, 0],
    ...         [3, 4, np.nan, 1],
    ...         [np.nan, np.nan, np.nan, 5],
    ...         [np.nan, 3, np.nan, 4],
    ...     ],
    ...     columns=list('ABCD'),
    ... )
    >>> df
        A    B   C  D
    0  NaN  2.0 NaN  0
    1  3.0  4.0 NaN  1
    2  NaN  NaN NaN  5
    3  NaN  3.0 NaN  4

    Replace all NaN elements with 0s.

    >>> tf = FillnaTF(0)
    >>> tf.transform(df)
        A   B   C   D
    0   0.0 2.0 0.0 0
    1   3.0 4.0 0.0 1
    2   0.0 0.0 0.0 5
    3   0.0 3.0 0.0 4

    We can also propagate non-null values forward or backward.

    >>> tf = FillnaTF(method='ffill')
    >>> tf.transform(df)
        A   B   C   D
    0   NaN 2.0 NaN 0
    1   3.0 4.0 NaN 1
    2   3.0 4.0 NaN 5
    3   3.0 3.0 NaN 4

    Replace all NaN elements in column 'A', 'B', 'C', and 'D', with 0, 1,
    2, and 3 respectively.

    >>> values = {'A': 0, 'B': 1, 'C': 2, 'D': 3}
    >>> tf = FillnaTF(value=values)
    >>> tf.transform(df)
        A   B   C   D
    0   0.0 2.0 2.0 0
    1   3.0 4.0 2.0 1
    2   0.0 1.0 2.0 5
    3   0.0 3.0 2.0 4

    Only replace the first NaN element.

    >>> tf = FillnaTF(value=values, limit=1)
    >>> tf.transform(df)
        A   B   C   D
    0   0.0 2.0 2.0 0
    1   3.0 4.0 NaN 1
    2   NaN 1.0 NaN 5
    3   NaN 3.0 NaN 4
    """

    transform_method = staticmethod(pd.DataFrame.fillna)



import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def groupby_index(s: pd.Series, /, **kwargs) -> pd.core.groupby.SeriesGroupBy:
    """
    Group Series by its index.

    Parameters
    ----------
    **kwargs
        See the documentation for :meth:`~pandas.Series.groupby` for complete
        details on the keyword arguments except ``by``.

    Returns
    -------
    SeriesGroupBy

    See Also
    --------
    pandas.Series.groupby
    dtoolkit.accessor.dataframe.groupby_index

    Notes
    -----
    ``by`` parameter is set to ``s.index`` and can't be overridden.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([1, 1, 1], index=['a', 'b', 'a'])
    >>> s
    a    1
    b    1
    a    1
    dtype: int64
    >>> s.groupby_index().count()
    a    2
    b    1
    dtype: int64
    """

    return s.groupby(s.index, **kwargs)



from __future__ import annotations

from typing import get_args
from typing import Literal

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas.base import GeometryArray
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.register import register_geoseries_method


CHINA_CRS = Literal["wgs84", "gcj02", "bd09"]


@register_geoseries_method
@doc(klass="GeoSeries")
def cncrs_offset(
    s: gpd.GeoSeries,
    /,
    from_crs: CHINA_CRS,
    to_crs: CHINA_CRS,
    a: float = 6378245,
    ee: float = 0.00669342162296594323,
) -> gpd.GeoSeries:
    r"""
    Fix the offset of the coordinates in China.

    Details see: `Restrictions on geographic data in China`__.

    __ https://en.wikipedia.org/wiki/Restrictions_on_geographic_data_in_China

    The following CRS could be transformed:

    - ``WGS-84`` to ``GCJ-02``.
    - ``WGS-84`` to ``BD-09``.
    - ``GCJ-02`` to ``WGS-84``.
    - ``GCJ-02`` to ``BD-09``.
    - ``BD-09`` to ``WGS-84``.
    - ``BD-09`` to ``GCJ-02``.

    Parameters
    ----------
    from_crs, to_crs : {{'wgs84', 'gcj02', 'bd09'}}
        The CRS of the input and output.

    a : float, default 6378245
        Semi major axis of the earth.

    ee : float, default 0.00669342162296594323
        Eccentricity\ :sup:`2`.

    Returns
    -------
    {klass}
        Replaced original geometry.

    Raises
    ------
    ValueError
        If the CRS is not ``ESGP:4326``.

    See Also
    --------
    geopandas.{klass}.to_crs
    dtoolkit.geoaccessor.geoseries.cncrs_offset
    dtoolkit.geoaccessor.geodataframe.cncrs_offset

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import pandas as pd
    >>> df = pd.DataFrame(
    ...     {{
    ...         "x": [114.21892734521, 128.543, 1],
    ...         "y": [29.575429778924, 37.065, 1],
    ...     }},
    ... ).from_xy("x", "y", crs=4326)
    >>> df
                x         y                    geometry
    0  114.218927  29.57543  POINT (114.21893 29.57543)
    1  128.543000  37.06500  POINT (128.54300 37.06500)
    2    1.000000   1.00000     POINT (1.00000 1.00000)
    >>> df.cncrs_offset(from_crs="bd09", to_crs="gcj02")
                x         y                    geometry
    0  114.218927  29.57543  POINT (114.21243 29.56938)
    1  128.543000  37.06500  POINT (128.53659 37.05875)
    2    1.000000   1.00000     POINT (1.00000 1.00000)
    """
    if s.crs != 4326:
        raise ValueError(f"Only support 'EPSG:4326' CRS, but got {s.crs!r}.")
    if from_crs == to_crs:
        raise ValueError("'from_crs' and 'to_crs' must be different.")
    elif from_crs not in get_args(CHINA_CRS):
        raise ValueError(
            f"Unknown 'from_crs': {from_crs!r}, must be in {get_args(CHINA_CRS)!r}.",
        )
    elif to_crs not in get_args(CHINA_CRS):
        raise ValueError(
            f"Unknown 'to_crs': {to_crs!r}, must be in {get_args(CHINA_CRS)!r}.",
        )

    s = s.copy()
    mask = (s.geom_type == "Point") & (is_in_china(s))
    if from_crs == "wgs84" and to_crs == "gcj02":
        s[mask] = wgs84_to_gcj02(s[mask], a=a, ee=ee)
    elif from_crs == "wgs84" and to_crs == "bd09":
        s[mask] = wgs84_to_bd09(s[mask], a=a, ee=ee)
    elif from_crs == "gcj02" and to_crs == "wgs84":
        s[mask] = gcj02_to_wgs84(s[mask], a=a, ee=ee)
    elif from_crs == "gcj02" and to_crs == "bd09":
        s[mask] = gcj02_to_bd09(s[mask])
    elif from_crs == "bd09" and to_crs == "wgs84":
        s[mask] = bd09_to_wgs84(s[mask], a=a, ee=ee)
    elif from_crs == "bd09" and to_crs == "gcj02":
        s[mask] = bd09_to_gcj02(s[mask])

    return s


def is_in_china(s: gpd.GeoSeries, /) -> pd.Series:
    """
    Based on China boundary to judge whether a point is in China.

    Parameters
    ----------
    GeoSeries
        The ESGP:4326 coordinates of the point.

    Returns
    -------
    Series of bool
    """
    from shapely.geometry import box

    return s.covered_by(box(73.66, 3.86, 135.05, 53.55))


# based on https://github.com/wandergis/coordTransform_py
def wgs84_to_gcj02(
    s: gpd.GeoSeries | GeometryArray,
    /,
    a: float,
    ee: float,
) -> GeometryArray:
    rad_y = s.y / 180 * np.pi
    magic = np.sqrt(1 - ee * np.sin(rad_y) ** 2)

    dx = transform_x(s) * 180 / (a / magic * np.cos(rad_y) * np.pi)
    dy = transform_y(s) * 180 / (a * (1 - ee) / magic**3 * np.pi)
    return gpd.points_from_xy(
        x=s.x + dx,
        y=s.y + dy,
    )


def wgs84_to_bd09(
    s: gpd.GeoSeries | GeometryArray,
    /,
    a: float,
    ee: float,
) -> GeometryArray:
    return gcj02_to_bd09(wgs84_to_gcj02(s, a=a, ee=ee))


# based on https://github.com/wandergis/coordTransform_py
def gcj02_to_wgs84(
    s: gpd.GeoSeries | GeometryArray,
    /,
    a: float,
    ee: float,
) -> GeometryArray:
    rad_y = s.y / 180 * np.pi
    magic = np.sqrt(1 - ee * np.sin(rad_y) ** 2)

    dx = transform_x(s) * 180 / (a / magic * np.cos(rad_y) * np.pi)
    dy = transform_y(s) * 180 / ((a * (1 - ee)) / magic**3 * np.pi)
    return gpd.points_from_xy(
        x=s.x - dx,
        y=s.y - dy,
    )


# based on https://github.com/wandergis/coordTransform_py
def transform_x(s: gpd.GeoSeries | GeometryArray, /) -> pd.Series:
    x, y = s.x - 105, s.y - 35
    x_dot_pi = x * np.pi

    return (
        300
        + x
        + 2 * y
        + 0.1 * x**2
        + 0.1 * x * y
        + 0.1 * np.sqrt(np.fabs(x))
        + (20 * np.sin(x_dot_pi * 6) + 20 * np.sin(x_dot_pi * 2)) * 2 / 3
        + (20 * np.sin(x_dot_pi) + 40 * np.sin(x_dot_pi / 3)) * 2 / 3
        + (150 * np.sin(x_dot_pi / 12) + 300 * np.sin(x_dot_pi / 30)) * 2 / 3
    )


# based on https://github.com/wandergis/coordTransform_py
def transform_y(s: gpd.GeoSeries | GeometryArray, /) -> pd.Series:
    x, y = s.x - 105, s.y - 35
    x_dot_pi, y_dot_pi = x * np.pi, y * np.pi

    return (
        -100
        + 2 * x
        + 3 * y
        + 0.2 * y**2
        + 0.1 * x * y
        + 0.2 * np.sqrt(np.fabs(x))
        + (20 * np.sin(x_dot_pi * 6) + 20 * np.sin(x_dot_pi * 2)) * 2 / 3
        + (20 * np.sin(y_dot_pi) + 40 * np.sin(y_dot_pi / 3)) * 2 / 3
        + (160 * np.sin(y_dot_pi / 12) + 320 * np.sin(y_dot_pi / 30)) * 2 / 3
    )


# based on https://github.com/wandergis/coordTransform_py
def gcj02_to_bd09(s: gpd.GeoSeries | GeometryArray, /) -> GeometryArray:
    PI = np.pi * 3000 / 180
    z = np.sqrt(s.x**2 + s.y**2) + 2e-5 * np.sin(s.y * PI)

    theta = np.arctan2(s.y, s.x) + 3e-6 * np.cos(s.x * PI)
    return gpd.points_from_xy(
        x=z * np.cos(theta) + 0.0065,
        y=z * np.sin(theta) + 0.006,
    )


def bd09_to_wgs84(
    s: gpd.GeoSeries | GeometryArray,
    /,
    a: float,
    ee: float,
) -> GeometryArray:
    return gcj02_to_wgs84(bd09_to_gcj02(s), a=a, ee=ee)


# based on https://github.com/wandergis/coordTransform_py
def bd09_to_gcj02(s: gpd.GeoSeries, /) -> GeometryArray:
    PI = np.pi * 3000 / 180
    x, y = s.x - 0.0065, s.y - 0.006
    z = np.sqrt(x**2 + y**2) - 2e-5 * np.sin(y * PI)

    theta = np.arctan2(y, x) - 3e-6 * np.cos(x * PI)
    return gpd.points_from_xy(
        x=z * np.cos(theta),
        y=z * np.sin(theta),
    )



from typing import Literal

import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.accessor.register import register_series_method


@register_series_method
@doc(klass="Series")
def dropna_index(s: pd.Series, /, how: Literal["any", "all"] = "any") -> pd.Series:
    """
    Return a new {klass} with missing index removed.

    Parameters
    ----------
    how : {{"any", "all"}}, default "any"
        If the ``Index`` is a :class:`~pandas.MultiIndex`, drop the value when "any"
        or "all" levels are NaN.

    Returns
    -------
    {klass}

    Raises
    ------
    ValueError
        If ``how`` isn't "any" or "all".

    See Also
    --------
    pandas.{klass}.dropna
    dtoolkit.accessor.series.dropna_index
    dtoolkit.accessor.dataframe.dropna_index

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> df = pd.DataFrame({{"a": [1, 2], "b": [3, 4]}}, index=[None, 0])
    >>> df
         a  b
    NaN  1  3
    0.0  2  4
    >>> df.dropna_index()
         a  b
    0.0  2  4
    """

    if how not in {"any", "all"}:
        raise ValueError(f"invalid how option: {how!r}")

    return s[~s.index._isnan] if s.index.hasnans else s



from __future__ import annotations

from functools import lru_cache
from functools import wraps
from typing import Callable
from warnings import warn

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like
from pandas.api.types import is_string_dtype

from dtoolkit.accessor.register import register_series_method
from dtoolkit.util._exception import find_stack_level


@register_series_method
def textdistance(
    s: pd.Series,
    /,
    other: None | str | pd.Series = None,
    method: Callable = None,
    align: bool = True,
) -> pd.Series:
    """
    Return a ``Series`` containing the text distance to aligned ``other``.

    Parameters
    ----------
    other : None, str or Series
        If None or nan, the text distance is 0.

    align : bool, default True
        If True, automatically aligns GeoSeries based on their indices. If False,
        the order of elements is preserved.

    method : Callable, default None
        The method to calculate the distance. The first and second positional parameters
        will be compared. If None, \
`rapidfuzz.fuzz.ratio <https://maxbachmann.github.io/RapidFuzz/Usage/fuzz.html#ratio>`_.

    Returns
    -------
    Series
        The values are the text distances.

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'rapidfuzz'.

    TypeError
        - If ``s`` is not string dtype.
        - If ``other`` is not string dtype.

    ValueError
        If ``other``'s length is not equal to the length of ``s``.

    See Also
    --------
    textdistance_matrix

    Notes
    -----
    The distance of any value compared to nan or None is 0.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series(["hello", "world"])
    >>> s
    0    hello
    1    world
    dtype: object
    >>> s.textdistance("python")
    0    36.363636
    1    18.181818
    dtype: float64
    >>> s.textdistance(pd.Series(["hello", "python"]))
    0    100.000000
    1     18.181818
    dtype: float64
    """

    if not is_string_dtype(s):
        raise TypeError(f"Expected string dtype, but got {s.dtype!r}.")

    if method is None:
        method = __import__("rapidfuzz.fuzz").fuzz.ratio
    method = lru_cache(check_none(check_nan(method)))

    if isinstance(other, str):
        return s.apply(method, args=(other,))
    elif isinstance(other, pd.Series):
        if not is_string_dtype(other):
            raise TypeError(f"Expected Series(string), but got {other.dtype!r}.")

        if align and not s.index.equals(other.index):
            warn("The indices are different.", stacklevel=find_stack_level())
            s, other = s.align(other)

        if s.size != other.size:
            raise ValueError(f"{s.size=} != {other.size=}.")

        return pd.Series(
            (method(*xy) for xy in zip(s, other)),
            name=s.name,
            index=s.index,
        )
    elif other is None or (not is_list_like(other) and pd.isna(other)):
        # NOTE:
        # - pd.na(Series) returns array-like of bool
        #   to make sure pd.isna(other) returns bool
        #   need to other is not array-like
        # - compare to None or nan always returns 0
        #   the behavior is following rapidfuzz.fuzz.ratio
        return pd.Series(np.zeros(s.size), name=s.name, index=s.index)

    raise TypeError(f"Expected Series(string), but got {type(other).__name__!r}.")


def check_none(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        # NOTE: compare to None always returns 0
        # the behavior is following rapidfuzz.fuzz.ratio
        return 0 if args[0] is None or args[1] is None else func(*args, **kwargs)

    return decorator


def check_nan(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        # NOTE: compare to nan always returns 0
        # the behavior is following rapidfuzz.fuzz.ratio
        return 0 if pd.isna(args[0]) or pd.isna(args[1]) else func(*args, **kwargs)

    return decorator



import pandas as pd

from dtoolkit.transformer.base import DataFrameTF


class QueryTF(DataFrameTF):
    """
    A transformer query the columns of a :obj:`~pandas.DataFrame` with a
    boolean expression.

    See Also
    --------
    pandas.DataFrame.query : This transformer's prototype method.

    Notes
    -----
    :meth:`~pandas.DataFrame.query`'s ``inplace`` parameter is not work for
    this transformer. Actually this break pipeline stream. If a transformer's
    ``inplace`` is ``True``, the next tf input would get ``None``.

    Examples
    --------
    >>> import pandas as pd
    >>> from dtoolkit.transformer import QueryTF
    >>> df = pd.DataFrame(
    ...     {
    ...         'A': range(1, 6),
    ...         'B': range(10, 0, -2),
    ...         'C C': range(10, 5, -1),
    ...     },
    ... )
    >>> df
        A   B  C C
    0  1  10   10
    1  2   8    9
    2  3   6    8
    3  4   4    7
    4  5   2    6
    >>> tf = QueryTF('A > B')
    >>> tf.transform(df)
        A  B  C C
    4  5  2    6

    The previous expression is equivalent to

    >>> df[df.A > df.B]
        A  B  C C
    4  5  2    6

    For columns with spaces in their name, you can use backtick quoting.

    >>> tf = QueryTF('B == `C C`')
    >>> tf.transform(df)
        A   B  C C
    0  1  10   10

    The previous expression is equivalent to

    >>> df[df.B == df['C C']]
        A   B  C C
    0  1  10   10
    """

    transform_method = staticmethod(pd.DataFrame.query)



from typing import Hashable

import geopandas as gpd
import pandas as pd

from dtoolkit._typing import SeriesOrFrame
from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
def xy(
    s: gpd.GeoSeries,
    /,
    reverse: bool = False,
    frame: bool = False,
    x: Hashable = "x",
    y: Hashable = "y",
) -> SeriesOrFrame:
    """
    Return the x and y location of Point geometries in a GeoSeries.

    Parameters
    ----------
    reverse : bool, default False
        If True, return (y, x) instead.

    frame : bool, default False
        If True, return a DataFrame instead of a Series.

    x : str, default 'x'
        Name of the x column if frame=True.

    y : str, default 'y'
        Name of the y column if frame=True.

    Returns
    -------
    Series or DataFrame
        If frame=False, a Series with tuple of coordinate. else, a DataFrame with
        x and y two columns.

    See Also
    --------
    geopandas.GeoSeries.x
    geopandas.GeoSeries.y

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> from shapely.geometry import Point
    >>> s = gpd.GeoSeries([Point(0, 1), Point(0, 2), Point(0, 3)])
    >>> s
    0    POINT (0.00000 1.00000)
    1    POINT (0.00000 2.00000)
    2    POINT (0.00000 3.00000)
    dtype: geometry

    Get the x and y coordinates of each point as a tuple.

    >>> s.xy()
    0    (0.0, 1.0)
    1    (0.0, 2.0)
    2    (0.0, 3.0)
    dtype: object

    Set ``reverse=True`` to return (y, x).

    >>> s.xy(True)
    0    (1.0, 0.0)
    1    (2.0, 0.0)
    2    (3.0, 0.0)
    dtype: object

    Set ``frame=True`` to return a DataFrame with x and y columns.

    >>> s.xy(frame=True)
         x    y
    0  0.0  1.0
    1  0.0  2.0
    2  0.0  3.0
    """

    coords = pd.concat((s.x.rename(x), s.y.rename(y)), axis=1)

    if reverse:
        coords = coords.iloc[:, ::-1]

    return coords if frame else coords.apply(tuple, axis=1)



from __future__ import annotations

from typing import Hashable

import geopandas as gpd
from pandas.util._decorators import doc

from dtoolkit._typing import Number
from dtoolkit._typing import OneDimArray
from dtoolkit.geoaccessor.geoseries import geobuffer as s_geobuffer
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_geobuffer, klass="GeoDataFrame", alias="df")
def geobuffer(
    df: gpd.GeoDataFrame,
    distance: Hashable | Number | list[Number] | OneDimArray,
    /,
    **kwargs,
) -> gpd.GeoDataFrame:

    # NOTE: Require pandas >= 1.3.0 to support `isinstance(SeriesOrDataFrame, Hashable)`
    if isinstance(distance, Hashable) and distance in df.columns:
        distance = df[distance]

    return df.assign(
        **{
            df.geometry.name: s_geobuffer(
                df.geometry,
                distance,
                **kwargs,
            ),
        }
    )



from dtoolkit.geoaccessor.series.from_wkb import from_wkb  # noqa: F401
from dtoolkit.geoaccessor.series.from_wkt import from_wkt  # noqa: F401
from dtoolkit.geoaccessor.series.geocode import geocode  # noqa: F401
from dtoolkit.geoaccessor.series.to_geoframe import to_geoframe  # noqa: F401
from dtoolkit.geoaccessor.series.to_geoseries import to_geoseries  # noqa: F401



import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def swap_index_values(s: pd.Series, /) -> pd.Series:
    """
    Swap the index and values of the Series.

    A sugary syntax wraps::

        pd.Series(s.index, index=s)

    Returns
    -------
    Series

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series(
    ...     [1, 2, 3],
    ...     index=pd.Index(
    ...         ["a", "b", "c"],
    ...         name='index',
    ...     ),
    ...     name='values',
    ... )
    >>> s
    index
    a    1
    b    2
    c    3
    Name: values, dtype: int64
    >>> s.swap_index_values()
    values
    1    a
    2    b
    3    c
    Name: index, dtype: object
    """

    return pd.Series(s.index, index=s)



import geopandas as gpd
from pandas.core.accessor import _register_accessor
from pandas.util._decorators import doc


@doc(klass=":class:`~geopandas.GeoSeries`")
def register_geoseries_accessor(name: str):
    """
    Register a custom accessor on {klass} objects.

    This is a temparatory solution to hook method into
    :class:`~geopandas.GeoSeries` or :class:`~geopandas.GeoDataFrame`.
    If `geopandas#1952`_ done, it would be removed from
    :mod:`dtoolkit.geoaccessor`.

    .. _geopandas#1952: https://github.com/geopandas/geopandas/pull/1952

    Parameters
    ----------
    name : str
        Name under which the accessor should be registered. A warning is issued
        if this name conflicts with a preexisting attribute.

    Returns
    -------
    callable
        A class decorator.

    See Also
    --------
    register_geoseries_accessor
    register_geodataframe_accessor
    dtoolkit.geoaccessor.register_geoseries_method
    dtoolkit.geoaccessor.register_geodataframe_method

    Notes
    -----
    When accessed, your accessor will be initialized with the geopandas object
    the user is interacting with. So the signature must be::

        def __init__(self, geopandas_object):  # noqa: E999
            ...

    For consistency with geopandas methods, you should raise an
    ``AttributeError`` if the data passed to your accessor has an incorrect
    dtype.

    >>> import geopandas as gpd
    >>> gpd.GeoSeries().dt
    Traceback (most recent call last):
    ...
    AttributeError: Can only use .dt accessor with datetimelike values

    Examples
    --------
    In your library code::

        from __future__ import annotations

        from dataclasses import dataclass

        import pandas as pd


        @register_geodataframe_accessor("gtype")
        @register_geoseries_accessor("gtype")
        @dataclass
        class GeoAccessor:

            _obj: gpd.GeoSeries | gpd.GeoDataFrame

            @property
            def is_point(self) -> pd.Series:
                # Return a boolean Series denoting whether each geometry is a point.

                return self._obj.geometry.geom_type == "Point"

    Back in an interactive IPython session:

    .. code-block:: ipython

        In [1]: import geopandas as gpd

        In [2]: s = gpd.GeoSeries.from_wkt(["POINT (0 0)", "POINT (1 1)", None])

        In [3]: s
        Out[3]:
        0    POINT (0.00000 0.00000)
        1    POINT (1.00000 1.00000)
        2                       None
        dtype: geometry

        In [4]: s.gtype.is_point
        Out[4]:
        0     True
        1     True
        2    False
        dtype: bool

        In [5]: d = s.to_frame("geometry")
        Out[5]:
                          geometry
        0  POINT (0.00000 0.00000)
        1  POINT (1.00000 1.00000)
        2                     None

        In [6]: d.gtype.is_point
        Out[6]:
        0     True
        1     True
        2    False
        dtype: bool
    """

    return _register_accessor(name, gpd.GeoSeries)


@doc(register_geoseries_accessor, klass=":class:`~geopandas.GeoDataFrame`")
def register_geodataframe_accessor(name: str):

    return _register_accessor(name, gpd.GeoDataFrame)



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.register import register_geoseries_method


@register_geoseries_method
@doc(klass="GeoSeries")
def get_coordinates(s: gpd.GeoSeries, /, **kwargs) -> pd.Series:
    """
    Gets coordinates from each geometry of :class:`~geopandas.{klass}`.

    A sugary syntax wraps :meth:`pygeos.coordinates.get_coordinates`.

    Parameters
    ----------
    **kwargs
        See the documentation for :meth:`pygeos.coordinates.get_coordinates` for
        complete details on the keyword arguments.

    Returns
    -------
    Series

    Raises
    ------
    ModuleNotFoundError
        If don't have module named 'pygeos'.

    See Also
    --------
    pygeos.coordinates.get_coordinates
        The core algorithm of this accessor.

    dtoolkit.geoaccessor.geoseries.get_coordinates
        Gets coordinates from each geometry of GeoSeries.

    dtoolkit.geoaccessor.geodataframe.get_coordinates
        Gets coordinates from each geometry of GeoDataFrame.

    Examples
    --------
    >>> import dtoolkit.geoaccessor
    >>> import geopandas as gpd
    >>> df = gpd.GeoSeries.from_wkt(
    ...     [
    ...         "POINT (0 0)",
    ...         "LINESTRING (2 2, 4 4)",
    ...         None,
    ...     ],
    ... ).to_frame("geometry")
    >>> df
                                            geometry
    0                        POINT (0.00000 0.00000)
    1  LINESTRING (2.00000 2.00000, 4.00000 4.00000)
    2                                           None
    >>> df.get_coordinates()
    0                [[0.0, 0.0]]
    1    [[2.0, 2.0], [4.0, 4.0]]
    2                          []
    Name: geometry, dtype: object
    """

    try:
        from shapely import get_coordinates

        # NOTE: requires shapely>=2.0

        return s.apply(get_coordinates)

    except ImportError:
        # TODO: delete pygeos after shapely 2.x released
        from pygeos import from_shapely, get_coordinates

        return s.apply(lambda x: get_coordinates(from_shapely(x), **kwargs))



import pandas as pd

from dtoolkit.accessor.register import register_series_method


@register_series_method
def to_set(s: pd.Series, /) -> set:
    """
    Return a :keyword:`set` of the values.

    A sugary syntax wraps :keyword:`set`::

        set(s)

    Returns
    -------
    set

    See Also
    --------
    pandas.Series.unique
    dtoolkit.accessor.index.to_set

    Notes
    -----
    Different to :meth:`~pandas.Index.unique`, it returns :class:`~pandas.Index`.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series([1, 2, 2])
    >>> s
    0    1
    1    2
    2    2
    dtype: int64
    >>> s.to_set()
    {1, 2}
    """

    return set(s.to_list())



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import (
    duplicated_geometry_groups as s_duplicated_geometry_groups,
)
from dtoolkit.geoaccessor.geoseries.duplicated_geometry_groups import BINARY_PREDICATE
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_duplicated_geometry_groups)
def duplicated_geometry_groups(
    df: gpd.GeoDataFrame,
    /,
    predicate: BINARY_PREDICATE = "intersects",
) -> pd.Series:
    return s_duplicated_geometry_groups(df.geometry, predicate=predicate)



import geopandas as gpd
import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.geoaccessor.geoseries import geoarea as s_geoarea
from dtoolkit.geoaccessor.register import register_geodataframe_method


@register_geodataframe_method
@doc(s_geoarea, alias="df")
def geoarea(df: gpd.GeoDataFrame, /) -> pd.Series:

    return s_geoarea(df.geometry)



from __future__ import annotations

import pandas as pd

from dtoolkit.accessor.register import register_series_method
from dtoolkit.accessor.series.dropna_index import dropna_index


@register_series_method
def values_to_dict(
    s: pd.Series,
    /,
    unique: bool = True,
    to_list: bool = True,
    dropna: bool = True,
) -> dict:
    """
    Convert :attr:`~pandas.Series.index` and :attr:`~pandas.Series.values` to
    :class:`dict`.

    Parameters
    ----------
    unique : bool, default True
        If True would drop duplicate elements.

    to_list : bool, default True
        If True one element value will return :class:`list`.

    dropna : bool, default True
        If True it will drop the ``nan`` value whatever it's key or value.

    Returns
    -------
    dict
        The datastruct is ``{index: [values]}``.

    See Also
    --------
    pandas.Series.to_dict
    dtoolkit.accessor.dataframe.values_to_dict

    Notes
    -----
    The same key of values would be merged into :class:`list`.

    Examples
    --------
    >>> import json
    >>> import dtoolkit
    >>> import pandas as pd
    >>> s = pd.Series(range(4), index=["a", "b", "a", "c"])
    >>> s
    a    0
    b    1
    a    2
    c    3
    dtype: int64
    >>> print(json.dumps(s.values_to_dict(), indent=4))
    {
        "a": [
            0,
            2
        ],
        "b": [
            1
        ],
        "c": [
            3
        ]
    }

    Unpack one element value list.

    >>> print(json.dumps(s.values_to_dict(to_list=False), indent=4))
    {
        "a": [
            0,
            2
        ],
        "b": 1,
        "c": 3
    }
    """

    if dropna:
        # Drop NA both index and values
        s = dropna_index(s.dropna())

    if s.empty:
        return {}

    return {
        key: handle_element(s[s.index == key], unique=unique, to_list=to_list)
        for key in s.index.unique()
    }


def handle_element(s: pd.Series, /, unique: bool, to_list: bool):
    if unique:
        s = s.unique()

    s = s.tolist()
    # Unfold one element list-like
    return s[0] if not to_list and len(s) == 1 else s



from dtoolkit.geoaccessor import dataframe  # noqa: F401
from dtoolkit.geoaccessor import geodataframe  # noqa: F401
from dtoolkit.geoaccessor import geoseries  # noqa: F401
from dtoolkit.geoaccessor import series  # noqa: F401
from dtoolkit.geoaccessor.accessor import register_geodataframe_accessor  # noqa: F401
from dtoolkit.geoaccessor.accessor import register_geoseries_accessor  # noqa: F401
from dtoolkit.geoaccessor.register import register_geodataframe_method  # noqa: F401
from dtoolkit.geoaccessor.register import register_geoseries_method  # noqa: F401



import pandas as pd

from dtoolkit.transformer.base import DataFrameTF


class ReplaceTF(DataFrameTF):
    """
    A transformer could replace values given input parameters.

    See Also
    --------
    pandas.DataFrame.replace : This transformer's prototype method.

    Notes
    -----
    :meth:`~pandas.DataFrame.replace`'s ``inplace`` parameter is not work for
    this transformer. Actually this break pipeline stream. If a transformer's
    ``inplace`` is ``True``, the next tf input would get ``None``.

    Examples
    --------

    Scalar ``to_replace`` and ``value``

    >>> import pandas as pd
    >>> from dtoolkit.transformer import ReplaceTF
    >>> df = pd.DataFrame(
    ...     {
    ...         'A': [0, 1, 2, 3, 4],
    ...         'B': [5, 6, 7, 8, 9],
    ...         'C': ['a', 'b', 'c', 'd', 'e'],
    ...     },
    ... )
    >>> tf = ReplaceTF(0, 5)
    >>> tf.transform(df)
    A  B  C
    0  5  5  a
    1  1  6  b
    2  2  7  c
    3  3  8  d
    4  4  9  e

    List-like ``to_replace``

    >>> tf = ReplaceTF([0, 1, 2, 3], 4)
    >>> tf.transform(df)
    A  B  C
    0  4  5  a
    1  4  6  b
    2  4  7  c
    3  4  8  d
    4  4  9  e

    >>> tf = ReplaceTF([0, 1, 2, 3], [4, 3, 2, 1])
    >>> tf.transform(df)
    A  B  C
    0  4  5  a
    1  3  6  b
    2  2  7  c
    3  1  8  d
    4  4  9  e

    dict-like ``to_replace``

    >>> tf = ReplaceTF({0: 10, 1: 100})
    >>> tf.transform(df)
        A  B  C
    0   10  5  a
    1  100  6  b
    2    2  7  c
    3    3  8  d
    4    4  9  e

    >>> tf = ReplaceTF({'A': 0, 'B': 5}, 100)
    >>> tf.transform(df)
        A    B  C
    0  100  100  a
    1    1    6  b
    2    2    7  c
    3    3    8  d
    4    4    9  e

    >>> tf = ReplaceTF({'A': {0: 100, 4: 400}})
    >>> tf.transform(df)
        A  B  C
    0  100  5  a
    1    1  6  b
    2    2  7  c
    3    3  8  d
    4  400  9  e

    Regular expression ``to_replace``

    >>> df = pd.DataFrame(
    ...     {
    ...         'A': ['bat', 'foo', 'bait'],
    ...         'B': ['abc', 'bar', 'xyz'],
    ...     },
    ... )
    >>> tf = ReplaceTF(to_replace=r'^ba.$', value='new', regex=True)
    >>> tf.transform(df)
        A    B
    0   new  abc
    1   foo  new
    2  bait  xyz

    >>> tf = ReplaceTF({'A': r'^ba.$'}, {'A': 'new'}, regex=True)
    >>> tf.transform(df)
        A    B
    0   new  abc
    1   foo  bar
    2  bait  xyz

    >>> tf = ReplaceTF(regex=r'^ba.$', value='new')
    >>> tf.transform(df)
        A    B
    0   new  abc
    1   foo  new
    2  bait  xyz

    >>> tf = ReplaceTF(regex={r'^ba.$': 'new', 'foo': 'xyz'})
    >>> tf.transform(df)
        A    B
    0   new  abc
    1   xyz  new
    2  bait  xyz

    >>> tf = ReplaceTF(regex=[r'^ba.$', 'foo'], value='new')
    >>> tf.transform(df)
        A    B
    0   new  abc
    1   new  new
    2  bait  xyz
    """

    transform_method = staticmethod(pd.DataFrame.replace)



import dtoolkit.accessor.dataframe  # noqa: F401
import dtoolkit.accessor.index  # noqa: F401
import dtoolkit.accessor.series  # noqa: F401
from dtoolkit.accessor.register import register_dataframe_method  # noqa: F401
from dtoolkit.accessor.register import register_index_method  # noqa: F401
from dtoolkit.accessor.register import register_method_factory  # noqa: F401
from dtoolkit.accessor.register import register_series_method  # noqa: F401



from __future__ import annotations

from typing import Hashable

import pandas as pd

from dtoolkit._typing import SeriesOrFrame
from dtoolkit.accessor.register import register_dataframe_method


@register_dataframe_method
def to_series(
    df: pd.DataFrame,
    /,
    name: Hashable = None,
    index_column: Hashable = None,
    value_column: Hashable = None,
) -> SeriesOrFrame:
    """
    Convert :class:`~pandas.DataFrame` to :class:`~pandas.Series`.

    Parameters
    ----------
    name : Hashable, optional
        The new name of returned Series. If not set, would use the original name.

    index_column : Hashable, optional
        The Series's index.

    value_column : Hashable, optional
        The Series's value.

    Returns
    -------
    Series or DataFrame
        Series if ``df`` is one column DataFrame or ``value_column`` is set.

    Raises
    ------
    ValueError
        - If ``index_column`` is same to ``value_column``.
        - If ``index_column`` is not in the columns.
        - If ``value_column`` is not in the columns.

    Examples
    --------
    >>> import dtoolkit
    >>> import pandas as pd

    Convert one column DataFrame to Series.

    >>> df = pd.DataFrame(range(3))
    >>> df
       0
    0  0
    1  1
    2  2
    >>> df.to_series(name="new name")
    0    0
    1    1
    2    2
    Name: new name, dtype: int64

    Can't directly convert two or more columns DataFrame

    >>> df = pd.DataFrame({"a": range(3), "b": range(3, 6), "c": range(6, 9)})
    >>> df
       a  b  c
    0  0  3  6
    1  1  4  7
    2  2  5  8
    >>> df.to_series()
       a  b  c
    0  0  3  6
    1  1  4  7
    2  2  5  8

    Convert to Series via ``value_column``.
    A sugar syntax for ``df.get(value_column)``.

    >>> df.to_series(value_column="a")
    0  0
    1  1
    2  2
    Name: a, dtype: int64

    Convert to Series via ``index_column`` and ``value_column``.
    A sugar syntax for ``df.set_index(index_column).get(value_column)``.

    >>> df.to_series(index_column="b", value_column="c")
    b
    3    6
    4    7
    5    8
    Name: c, dtype: int64
    """
    if df.columns.size == 1:  # one column
        return _to_series(df, name=name, value_column=df.columns[0])

    elif value_column is not None:  # two or more columns
        if index_column == value_column:
            raise ValueError(
                f"'index_column' ({index_column}) and 'value_column' ({value_column}) "
                "should be different.",
            )

        if index_column is None:  # use original index
            return _to_series(df, name=name, value_column=value_column)

        # use `index_column` as index
        return _to_series(
            df.set_index(index_column),
            name=name,
            value_column=value_column,
        )

    return df


def _to_series(
    df: pd.DataFrame,
    /,
    name: Hashable,
    value_column: Hashable,
) -> pd.Series:
    """
    Select one column (`value_column`) of DataFrame and convert to Series.
    The name of Series is set to `name`.
    """

    return df[value_column].rename(name or value_column)



from typing import Literal

import pandas as pd
from pandas.util._decorators import doc

from dtoolkit.accessor.register import register_dataframe_method
from dtoolkit.accessor.series import dropna_index as s_dropna_index


@register_dataframe_method
@doc(s_dropna_index, klass="DataFrame")
def dropna_index(
    df: pd.DataFrame,
    /,
    how: Literal["any", "all"] = "any",
) -> pd.DataFrame:
    return s_dropna_index(df, how=how)
